import re
import numpy as np
from copy import deepcopy

import astropy.units as u
from specutils import Spectrum1D
from specutils.utils import QuantityModel
from traitlets import Bool, List, Unicode, observe

from jdaviz.configs.default.plugins.model_fitting.fitting_backend import fit_model_to_spectrum
from jdaviz.configs.default.plugins.model_fitting.initializers import (MODELS,
                                                                       initialize,
                                                                       get_model_parameters)
from jdaviz.core.events import SnackbarMessage, GlobalDisplayUnitChanged
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        SelectPluginComponent,
                                        SpectralSubsetSelectMixin,
                                        DatasetSelectMixin,
                                        DatasetSpectralSubsetValidMixin,
                                        NonFiniteUncertaintyMismatchMixin,
                                        AutoTextField,
                                        AddResultsMixin,
                                        TableMixin,
                                        with_spinner)
from jdaviz.core.custom_traitlets import IntHandleEmpty
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               flux_conversion_general)
from jdaviz.core.custom_units_and_equivs import PIX2

__all__ = ['ModelFitting']


class _EmptyParam:
    def __init__(self, value, unit=None):
        self.value = value
        self.unit = unit
        self.quantity = u.Quantity(self.value,
                                   self.unit if self.unit is not None else u.dimensionless_unscaled)


@tray_registry('g-model-fitting', label="Model Fitting", viewer_requirements='spectrum')
class ModelFitting(PluginTemplateMixin, DatasetSelectMixin,
                   SpectralSubsetSelectMixin, DatasetSpectralSubsetValidMixin,
                   NonFiniteUncertaintyMismatchMixin,
                   AddResultsMixin, TableMixin):
    """
    See the :ref:`Model Fitting Plugin Documentation <specviz-model-fitting>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``cube_fit``
      Only exposed for Cubeviz.  Whether to fit the model to the cube instead of to the
      collapsed spectrum.
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to fit the model.
    * ``spectral_subset`` (:class:`~jdaviz.core.template_mixin.SubsetSelect`)
    * ``model_component`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``poly_order``
    * ``model_component_label`` (:class:`~jdaviz.core.template_mixin.AutoTextField`)
    * :meth:`create_model_component`
    * :meth:`remove_model_component`
    * :meth:`model_components`
    * :meth:`valid_model_components`
    * :meth:`get_model_component`
    * :meth:`set_model_component`
    * :meth:`reestimate_model_parameters`
    * ``equation`` (:class:`~jdaviz.core.template_mixin.AutoTextField`)
    * :meth:`equation_components`
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * ``residuals_calculate`` (bool)
      Whether to calculate and expose the residuals (model minus data).
    * ``residuals`` (:class:`~jdaviz.core.template_mixin.AutoTextField`)
      Label of the residuals to apply when calling :meth:`calculate_fit` if ``residuals_calculate``
      is ``True``.
    * :meth:`calculate_fit`
    * :meth:`fitted_models`
    * :meth:`get_models`
    * :meth:`get_model_parameters`
    """
    dialog = Bool(False).tag(sync=True)
    template_file = __file__, "model_fitting.vue"
    form_valid_model_component = Bool(False).tag(sync=True)

    # model components:
    model_comp_items = List().tag(sync=True)
    model_comp_selected = Unicode().tag(sync=True)
    poly_order = IntHandleEmpty(0).tag(sync=True)

    comp_label = Unicode().tag(sync=True)
    comp_label_default = Unicode().tag(sync=True)
    comp_label_auto = Bool(True).tag(sync=True)
    comp_label_invalid_msg = Unicode().tag(sync=True)

    model_equation = Unicode().tag(sync=True)
    model_equation_default = Unicode().tag(sync=True)
    model_equation_auto = Bool(True).tag(sync=True)
    model_equation_invalid_msg = Unicode().tag(sync=True)

    eq_error = Bool(False).tag(sync=True)
    component_models = List([]).tag(sync=True)
    display_order = Bool(False).tag(sync=True)

    cube_fit = Bool(False).tag(sync=True)

    # residuals (non-cube fit only)
    residuals_calculate = Bool(False).tag(sync=True)
    residuals_label = Unicode().tag(sync=True)
    residuals_label_default = Unicode().tag(sync=True)
    residuals_label_auto = Bool(True).tag(sync=True)
    residuals_label_invalid_msg = Unicode('').tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._units = {}  # string representation of display units
        self._fitted_model = None
        self._fitted_spectrum = None
        self._fitted_models = {}
        self.component_models = []
        self._initialized_models = {}
        self._display_order = False

        # description displayed under plugin title in tray
        self._plugin_description = 'Fit an analytic model to data or a subset of data.'

        # create the label first so that when model_component defaults to the first selection,
        # the label automatically defaults as well
        self.model_component_label = AutoTextField(self, 'comp_label', 'comp_label_default',
                                                   'comp_label_auto', 'comp_label_invalid_msg')

        self.model_component = SelectPluginComponent(self,
                                                     items='model_comp_items',
                                                     selected='model_comp_selected',
                                                     manual_options=list(MODELS.keys()))

        if self.config == 'cubeviz':
            # use mean whenever extracting the 1D spectrum of a cube to initialize model params
            self.dataset._spectral_extraction_function = 'Mean'
        # by default, require entries to be in spectrum-viewer (not other cubeviz images, etc)
        # in cubeviz, the cube_fit toggle will then replace this filter to filter for cubes
        self.dataset.add_filter('layer_in_spectrum_viewer')

        self.equation = AutoTextField(self, 'model_equation', 'model_equation_default',
                                      'model_equation_auto', 'model_equation_invalid_msg')

        self.residuals = AutoTextField(self, 'residuals_label', 'residuals_label_default',
                                       'residuals_label_auto', 'residuals_label_invalid_msg')

        headers = ['model', 'data_label', 'spectral_subset', 'equation']
        if self.config == 'cubeviz':
            headers += ['cube_fit']

        self.table.headers_avail = headers
        self.table.headers_visible = headers
        # when model parameters are added as columns, only show the value/unit columns by default
        # (other columns can be show in the dropdown by the user)
        self.table._new_col_visible = lambda colname: colname.split(':')[-1] not in ('fixed', 'uncert', 'std')  # noqa

        # set the filter on the viewer options
        self._update_viewer_filters()

        self.hub.subscribe(self, GlobalDisplayUnitChanged,
                           handler=self._on_global_display_unit_changed)

    @property
    def _default_spectrum_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'spectrum-viewer'
        )

    @property
    def _default_flux_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_flux_viewer_reference_name', 'flux-viewer'
        )

    @property
    def user_api(self):
        expose = ['dataset']
        if self.config == "cubeviz":
            expose += ['cube_fit']
        expose += ['spectral_subset', 'model_component', 'poly_order', 'model_component_label',
                   'model_components', 'valid_model_components',
                   'create_model_component', 'remove_model_component',
                   'get_model_component', 'set_model_component', 'reestimate_model_parameters',
                   'equation', 'equation_components',
                   'add_results', 'residuals_calculate', 'residuals']
        expose += ['calculate_fit', 'clear_table', 'export_table',
                   'fitted_models', 'get_models', 'get_model_parameters']
        return PluginUserApi(self, expose=expose)

    def _param_units(self, param, model_type=None):
        """Helper function to handle units that depend on x and y"""
        y_params = ["amplitude", "amplitude_L", "intercept", "scale"]

        if param == "slope":
            return str(u.Unit(self._units["y"]) / u.Unit(self._units["x"]))
        elif model_type == 'Polynomial1D':
            # param names are all named cN, where N is the order
            order = int(float(param[1:]))
            return str(u.Unit(self._units["y"]) / u.Unit(self._units["x"])**order)
        elif param == "temperature":
            return str(u.K)
        elif param == "scale" and model_type == "BlackBody":
            return str("")

        return self._units["y"] if param in y_params else self._units["x"]

    def _update_parameters_from_fit(self):
        """Insert the results of the model fit into the component_models"""
        for m in self.component_models:
            submodel_index = None
            name = m["id"]
            if hasattr(self._fitted_model, "submodel_names"):
                for i in range(len(self._fitted_model.submodel_names)):
                    if name == self._fitted_model.submodel_names[i]:
                        m_fit = self._fitted_model[name]
                        submodel_index = i
                if submodel_index is None:
                    continue
            elif self._fitted_model.name == name:
                m_fit = self._fitted_model
            else:
                # then the component was not in the fitted model
                continue

            temp_params = []
            for i in range(0, len(m_fit.parameters)):
                temp_param = [x for x in m["parameters"] if x["name"] ==
                              m_fit.param_names[i]]
                temp_param[0]["value"] = m_fit.parameters[i]
                # The submodels don't have uncertainties attached, only the compound model
                if self._fitted_model.stds is not None:
                    std_name = temp_param[0]["name"]
                    if submodel_index is not None:
                        std_name = f"{std_name}_{submodel_index}"
                    if std_name in self._fitted_model.stds.param_names:
                        temp_param[0]["std"] = self._fitted_model.stds[std_name]

                temp_params += temp_param

            m["parameters"] = temp_params

        self.send_state('component_models')

    def _update_parameters_from_QM(self):
        """
        Parse out result parameters from a QuantityModel, which isn't
        subscriptable with model name
        """
        if hasattr(self._fitted_model, "submodel_names"):
            submodel_names = self._fitted_model.submodel_names
            submodels = True
        else:
            submodel_names = [self._fitted_model.name]
            submodels = False
        fit_params = self._fitted_model.parameters
        param_names = self._fitted_model.param_names

        for i in range(len(submodel_names)):
            name = submodel_names[i]
            m = [x for x in self.component_models if x["id"] == name][0]
            temp_params = []
            if submodels:
                idxs = [j for j in range(len(param_names)) if
                        int(param_names[j][-1]) == i]
            else:
                idxs = [j for j in range(len(param_names))]
            # This is complicated by needing to handle parameter names that
            # have underscores in them, since QuantityModel adds an underscore
            # and integer to indicate to which model a parameter belongs
            for idx in idxs:
                if submodels:
                    temp_param = [x for x in m["parameters"] if x["name"] ==
                                  "_".join(param_names[idx].split("_")[0:-1])]
                else:
                    temp_param = [x for x in m["parameters"] if x["name"] ==
                                  param_names[idx]]
                temp_param[0]["value"] = fit_params[idx]
                temp_params += temp_param
            m["parameters"] = temp_params

        self.send_state('component_models')

    def _update_initialized_parameters(self):
        # If the user changes a parameter value, we need to change it in the
        # initialized model
        for m in self.component_models:
            name = m["id"]
            for param in m["parameters"]:
                quant_param = u.Quantity(param["value"], param["unit"])
                setattr(self._initialized_models[name], param["name"],
                        quant_param)

    def _warn_if_no_equation(self):
        if self.model_equation == "" or self.model_equation is None:
            example = "+".join([m["id"] for m in self.component_models])
            snackbar_message = SnackbarMessage(
                f"Error: a model equation must be defined, e.g. {example}",
                color='error',
                sender=self)
            self.hub.broadcast(snackbar_message)
            return True
        else:
            return False

    def _update_viewer_filters(self, event={}):
        if event.get('new', self.cube_fit):
            # only want image viewers in the options
            self.add_results.viewer.filters = ['is_image_viewer']
        else:
            # only want spectral viewers in the options
            self.add_results.viewer.filters = ['is_spectrum_viewer']

    @observe('cube_fit')
    def _cube_fit_changed(self, event={}):
        self._update_viewer_filters(event=event)

        sb_unit = self.app._get_display_unit('sb')
        spectral_y_unit = self.app._get_display_unit('spectral_y')
        if event.get('new'):
            self._units['y'] = str(sb_unit)
            self.dataset.add_filter('is_flux_cube')
            self.dataset.remove_filter('layer_in_spectrum_viewer')
        else:
            self._units['y'] = str(spectral_y_unit)
            self.dataset.add_filter('layer_in_spectrum_viewer')
            self.dataset.remove_filter('is_flux_cube')

        self.dataset._clear_cache()
        if sb_unit != spectral_y_unit:
            # We make the user hit the reestimate button themselves
            for model_index, comp_model in enumerate(self.component_models):
                self.component_models[model_index]["compat_display_units"] = False
            self.send_state('component_models')

    @observe("dataset_selected")
    def _dataset_selected_changed(self, event=None):
        """
        Callback method for when the user has selected data from the drop down
        in the front-end. It is here that we actually parse and create a new
        data object from the selected data. From this data object, unit
        information is scraped, and the selected spectrum is stored for later
        use in fitting.

        Parameters
        ----------
        event : str
            IPyWidget callback event object. In this case, represents the data
            label of the data collection object selected by the user.
        """
        if not hasattr(self, 'dataset') or self.app._jdaviz_helper is None or self.dataset_selected == '':  # noqa
            # during initial init, this can trigger before the component is initialized
            return

        selected_spec = self.dataset.selected_obj
        if selected_spec is None:
            return

        # Replace NaNs from collapsed Spectrum1D in Cubeviz
        # (won't affect calculations because these locations are masked)
        selected_spec.flux[np.isnan(selected_spec.flux)] = 0.0

    def _default_comp_label(self, model, poly_order=None):
        abbrevs = {'BlackBody': 'BB', 'PowerLaw': 'PL', 'Lorentz1D': 'Lo'}
        abbrev = abbrevs.get(model, model[0].upper())
        if model == "Polynomial1D":
            abbrev += f'{poly_order}'

        # append a number suffix to avoid any duplicates
        ind = 1
        while abbrev in [cm['id'] for cm in self.component_models]:
            abbrev = f'{abbrev.split("_")[0]}_{ind}'
            ind += 1

        return abbrev

    @observe('model_comp_selected', 'poly_order')
    def _update_comp_label_default(self, event={}):
        self.display_order = self.model_comp_selected == "Polynomial1D"
        self.comp_label_default = self._default_comp_label(self.model_comp_selected,
                                                           self.poly_order)

    @observe('comp_label')
    def _comp_label_changed(self, event={}):
        if not len(self.comp_label.strip()):
            # strip will raise the same error for a label of all spaces
            self.comp_label_invalid_msg = 'label must be provided'
            return
        if self.comp_label in [cm['id'] for cm in self.component_models]:
            self.comp_label_invalid_msg = 'label already in use'
            return
        self.comp_label_invalid_msg = ''

    def _update_model_equation_default(self):
        self.model_equation_default = '+'.join(cm['id'] for cm in self.component_models)

    def _reinitialize_with_fixed(self):
        """
        Reinitialize all component models with current values and the
        specified parameters fixed (can't easily update fixed dictionary in
        an existing model)
        """
        temp_models = []
        for m in self.component_models:
            fixed = {}

            # Set the initial values as quantities to make sure model units
            # are set correctly.
            initial_values = {p["name"]: u.Quantity(p["value"], p["unit"]) for p in m["parameters"]}

            for p in m["parameters"]:
                fixed[p["name"]] = p["fixed"]

            # Have to initialize with fixed dictionary
            temp_model = MODELS[m["model_type"]](name=m["id"], fixed=fixed,
                                                 **initial_values, **m.get("model_kwargs", {}))

            temp_models.append(temp_model)

        return temp_models

    def create_model_component(self, model_component=None, model_component_label=None,
                               poly_order=None):
        """
        Add a new model component to the list of available model components

        Parameters
        ----------
        model_component : str
            Type of model component to add.  If not provided, will default according to
            ``model_component``.
        model_component_label : str
            Name of the model component to add.  If not provided, will default according to
            ``model_component_label`` (if ``model_component_label.auto`` is True and
            ``model_component`` is passed as an argument, then the default label will be recomputed
            rather than applying the current value).
        poly_order : int
            Order of the polynomial if ``model_component`` is (or defaults to) "Polynomial1D".
            Will raise an error if provided and ``model_component`` is not "Polynomial1D".
            If not provided, will default according to ``poly_order``.
        """
        model_comp = model_component if model_component is not None else self.model_comp_selected

        if model_comp != "Polynomial1D" and poly_order is not None:
            raise ValueError("poly_order should only be passed if model_component is Polynomial1D")
        poly_order = poly_order if poly_order is not None else self.poly_order

        # if model_component was passed and different than the one set in the traitlet, AND
        # model_component_label is not passed, AND the auto is enabled on the label, then
        # recompute a temporary default model label rather than use the value set in the traitlet
        if model_comp != self.model_comp_selected and model_component_label is None and self.model_component_label.auto:  # noqa
            comp_label = self._default_comp_label(model_comp, poly_order)
        else:
            comp_label = model_component_label if model_component_label is not None else self.comp_label  # noqa

        # validate provided label (only allow "word characters").   These should already be
        # stripped by JS in the UI element, but we'll confirm here (especially if this is ever
        # extended to have better API-support)
        if re.search(r'\W+', comp_label) or not len(comp_label):
            raise ValueError(f"invalid model component label '{comp_label}'")

        if comp_label in [cm['id'] for cm in self.component_models]:
            raise ValueError(f"model component label '{comp_label}' already in use")

        new_model = self._initialize_model_component(model_comp, comp_label, poly_order=poly_order)
        self.component_models = self.component_models + [new_model]
        # update the default label (likely adding the suffix)
        self._update_comp_label_default()
        self._update_model_equation_default()

    def _initialize_model_component(self, model_comp, comp_label, poly_order=None):
        new_model = {"id": comp_label, "model_type": model_comp,
                     "parameters": [], "model_kwargs": {}}
        model_cls = MODELS[model_comp]

        # Need to set the units the first time we initialize a model component, after this
        # we listen for display unit changes
        if self._units.get('x', '') == '':
            self._units['x'] = str(self.app._get_display_unit('spectral'))
        if self._units.get('y', '') == '':
            if self.cube_fit:
                self._units['y'] = str(self.app._get_display_unit('sb'))
            else:
                self._units['y'] = str(self.app._get_display_unit('spectral_y'))

        if model_comp == "Polynomial1D":
            # self.poly_order is the value in the widget for creating
            # the new model component.  We need to store that with the
            # model itself as the value could change for another component.
            new_model["model_kwargs"] = {"degree": poly_order}
        elif model_comp == "BlackBody":
            new_model["model_kwargs"] = {"output_units": self._units["y"],
                                         "bounds": {"scale": (0.0, None)}}

        initial_values = {}
        for param_name in get_model_parameters(model_cls, new_model["model_kwargs"]):
            # access the default value from the model class itself
            default_param = getattr(model_cls, param_name, _EmptyParam(0))
            default_units = self._param_units(param_name,
                                              model_type=new_model["model_type"])

            if default_param.unit is None:
                # then the model parameter accepts unitless, but we want
                # to pass with appropriate default units
                initial_val = u.Quantity(default_param.value, default_units)
            else:
                # then the model parameter has default units.  We want to pass
                # with jdaviz default units (based on x/y units) but need to
                # convert the default parameter unit to these units
                if default_param.unit != default_units:
                    pixar_sr = self.app.data_collection[0].meta.get('PIXAR_SR', 1)
                    viewer = self.app.get_viewer("spectrum-viewer")
                    cube_wave = viewer.slice_value * u.Unit(self.app._get_display_unit('spectral'))
                    equivs = all_flux_unit_conversion_equivs(pixar_sr, cube_wave)

                    initial_val = flux_conversion_general([default_param.value],
                                                          default_param.unit,
                                                          default_units, equivs)
                else:
                    initial_val = default_param

            initial_values[param_name] = initial_val

        if self.cube_fit:
            # We need to input the whole cube when initializing the model so the units are correct.
            if self.dataset_selected in self.app.data_collection.labels:
                data = self.app.data_collection[self.dataset_selected].get_object(statistic=None)
            else:  # User selected some subset from spectrum viewer, just use original cube
                data = self.app.data_collection[0].get_object(statistic=None)
            masked_spectrum = self._apply_subset_masks(data, self.spectral_subset)
        else:
            masked_spectrum = self._apply_subset_masks(self.dataset.selected_spectrum,
                                                       self.spectral_subset)
        mask = masked_spectrum.mask
        if mask is not None:
            if mask.ndim == 3:
                spectral_mask = mask.all(axis=(0, 1))
            else:
                spectral_mask = mask
            init_x = masked_spectrum.spectral_axis[~spectral_mask]
            orig_flux_shape = masked_spectrum.flux.shape
            init_y = masked_spectrum.flux[~mask]
            if mask.ndim == 3:
                init_y = init_y.reshape(orig_flux_shape[0],
                                        orig_flux_shape[1],
                                        len(init_x))
        else:
            init_x = masked_spectrum.spectral_axis
            init_y = masked_spectrum.flux

        if init_y.unit != self._units['y']:
            # equivs for spectral density and flux<>sb
            pixar_sr = masked_spectrum.meta.get('_pixel_scale_factor', 1.0)
            equivs = all_flux_unit_conversion_equivs(pixar_sr, init_x)

            init_y = flux_conversion_general([init_y.value],
                                             init_y.unit,
                                             self._units['y'],
                                             equivs)

        initialized_model = initialize(
            MODELS[model_comp](name=comp_label,
                               **initial_values,
                               **new_model.get("model_kwargs", {})),
            init_x, init_y)

        # need to loop over parameters again as the initializer may have overridden
        # the original default value. However, if we toggled cube_fit, we may need to override
        for param_name in get_model_parameters(model_cls, new_model["model_kwargs"]):
            param_quant = getattr(initialized_model, param_name)
            new_model["parameters"].append({"name": param_name,
                                            "value": param_quant.value,
                                            "unit": str(param_quant.unit),
                                            "fixed": False})
        self._initialized_models[comp_label] = initialized_model

        new_model["Initialized"] = True
        new_model["initialized_display_units"] = self._units.copy()

        new_model["compat_display_units"] = True  # always compatible at time of creation
        return new_model

    def _check_model_component_compat(self, axes=['x', 'y'], display_units=None):
        if display_units is None:
            display_units = [u.Unit(self._units[ax]) for ax in axes]

        disp_physical_types = [unit.physical_type for unit in display_units]

        for model_index, comp_model in enumerate(self.component_models):
            compat = True
            for ax, ax_physical_type in zip(axes, disp_physical_types):
                comp_unit = u.Unit(comp_model["initialized_display_units"][ax])
                compat = comp_unit.physical_type == ax_physical_type
                if not compat:
                    break
            self.component_models[model_index]["compat_display_units"] = compat

        # length hasn't changed, so we need to force the traitlet to update
        self.send_state("component_models")
        self._check_model_equation_invalid()

    def _on_global_display_unit_changed(self, msg):
        if msg.axis == 'spectral_y':
            axis = 'y'
        elif msg.axis == 'spectral':
            axis = 'x'
        else:
            return

        if axis == 'y' and self.cube_fit:
            # The units have to be in surface brightness for a cube fit.
            uc = self.app._jdaviz_helper.plugins['Unit Conversion']

            if msg.unit != uc._obj.sb_unit_selected:
                self._units[axis] = uc._obj.sb_unit_selected
                self._check_model_component_compat([axis], [u.Unit(uc._obj.sb_unit_selected)])
                return

        # update internal tracking of current units
        self._units[axis] = str(msg.unit)

        self._check_model_component_compat([axis], [msg.unit])

    def remove_model_component(self, model_component_label):
        """
        Remove an existing model component.

        Parameters
        ----------
        model_component_label : str
            The label given to the existing model component
        """
        if model_component_label not in [x["id"] for x in self.component_models]:
            raise ValueError(f"model component with label '{model_component_label}' does not exist")

        self.component_models = [x for x in self.component_models
                                 if x["id"] != model_component_label]
        del self._initialized_models[model_component_label]
        self._update_comp_label_default()
        self._update_model_equation_default()

    def get_model_component(self, model_component_label, parameter=None):
        """
        Get a (read-only) dictionary representation of an existing model component.

        Parameters
        ----------
        model_component_label : str
            The label given to the existing model component
        parameter : str
            Optional.  The name of a valid parameter in the model component, in which case only
            the information on that parameter is returned.
        """
        try:
            model_component = [x for x in self.component_models
                               if x["id"] == model_component_label][0]
        except IndexError:
            raise ValueError(f"'{model_component_label}' is not a label of an existing model component")  # noqa

        comp = {"model_type": model_component['model_type'],
                "parameters": {p['name']: {'value': p['value'],
                                           'unit': p['unit'],
                                           'std': p.get('std', np.nan),
                                           'fixed': p['fixed']} for p in model_component['parameters']}}  # noqa

        if parameter is not None:
            return comp['parameters'].get(parameter)
        return comp

    def set_model_component(self, model_component_label, parameter, value=None, fixed=None):
        """
        Set the value or fixed attribute of a parameter in an existing model component.

        Parameters
        ----------
        model_component_label : str
            The label given to the existing model component
        parameter : str
            The name of a valid parameter in the model component.
        value : float
            Optional.  The new initial value of the parameter.  If not provided or None, will
            remain unchanged.
        fixed : bool
            Optional.  The new state of the fixed attribute of the parameter.  If not provided
            or None, will remain unchanged.

        Returns
        -------
        updated dictionary of the parameter representation
        """
        cms = self.component_models
        try:
            model_component = [x for x in cms if x["id"] == model_component_label][0]
        except IndexError:
            raise ValueError(f"'{model_component_label}' is not a label of an existing model component")  # noqa
        try:
            parameter = [p for p in model_component['parameters'] if p['name'] == parameter][0]
        except IndexError:
            raise ValueError(f"'{parameter}' is not the name of a parameter in the '{model_component_label}' model component")  # noqa

        if value is not None:
            if not isinstance(value, (int, float)):
                raise TypeError("value must be a float")
            parameter['value'] = value
        if fixed is not None:
            if not isinstance(fixed, bool):
                raise TypeError("fixed must be a boolean")
            parameter['fixed'] = fixed

        self.component_models = []
        self.component_models = cms

        return parameter

    def vue_reestimate_model_parameters(self, model_component_label=None, **kwargs):
        self.reestimate_model_parameters(model_component_label=model_component_label)

    def reestimate_model_parameters(self, model_component_label=None):
        """
        Re-estimate all free parameters in a given model component given the currently selected
        data and subset selections.

        Parameters
        ----------
        model_component_label : str or None.
            The label given to the existing model component.  If None, will iterate over all model
            components.
        """
        if model_component_label is None:
            return [self.reestimate_model_parameters(model_comp["id"])
                    for model_comp in self.component_models]

        try:
            model_index, model_comp = [(i, x) for i, x in enumerate(self.component_models)
                                       if x["id"] == model_component_label][0]
        except IndexError:
            raise ValueError(f"'{model_component_label}' is not a label of an existing model component")  # noqa

        # store user-fixed parameters so we can revert after re-initializing
        fixed_params = {p['name']: p for p in model_comp['parameters'] if p['fixed']}

        new_model = self._initialize_model_component(model_comp['model_type'],
                                                     model_component_label,
                                                     poly_order=model_comp['model_kwargs'].get('degree', None))  # noqa

        # revert fixed parameters to user-value
        new_model['parameters'] = [fixed_params.get(p['name'], p) for p in new_model['parameters']]

        self.component_models[model_index] = new_model
        # length hasn't changed, so we need to force the traitlet to update
        self.send_state("component_models")

        # model units may have changed, need to re-check their compatibility with display units
        self._check_model_component_compat()

        # return user-friendly info on revised model
        return self.get_model_component(model_component_label)

    @property
    def model_components(self):
        """
        List of the labels of existing model components
        """
        return [x["id"] for x in self.component_models]

    @property
    def valid_model_components(self):
        """
        List of the labels of existing valid (due to display units) model components
        """
        return [x["id"] for x in self.component_models if x["compat_display_units"]]

    @property
    def equation_components(self):
        """
        List of the labels of model components in the current equation
        """
        return re.split(r'[+*/-]', self.equation.value.replace(' ', ''))

    @property
    def fitted_models(self):
        """
        Dictionary of all previously fitted models.
        """
        return self._fitted_models

    def get_models(self, models=None, model_label=None, x=None, y=None):
        """
        Loop through all models and output models of the label model_label.
        If x or y is set, return model_labels of those (x, y) coordinates.
        If x and y are None, print all models regardless of coordinates.

        Parameters
        ----------
        models : dict
            A dict of models, with the key being the label name and the value
            being an `astropy.modeling.CompoundModel` object. Defaults to
            `fitted_models` if no parameter is provided.
        model_label : str
            The name of the model that will be found and returned. If it
            equals default, every model present will be returned.
        x : int
            The x coordinate of the model spaxels that will be returned.
        y : int
            The y coordinate of the model spaxels that will be returned.

        Returns
        -------
        selected_models : dict
            Dictionary of the selected models.
        """
        selected_models = {}
        # If models is not provided, use the app's fitted models
        if not models:
            models = self.fitted_models

        # Loop through all keys in the dict models
        for label in models:
            # Prevent "Model 2" from being returned when model_label is "Model"
            if model_label is not None:
                if label.split(" (")[0] != model_label:
                    continue

            # If no label was provided, use label name without coordinates.
            if model_label is None and " (" in label:
                find_label = label.split(" (")[0]
            # If coordinates are not present, just use the label.
            elif model_label is None:
                find_label = label
            else:
                find_label = model_label

            # If x and y are set, return keys that match the model plus that
            # coordinate pair. If only x or y is set, return keys that fit
            # that value for the appropriate coordinate.
            if x is not None and y is not None:
                find_label = r"{} \({}, {}\)".format(find_label, x, y)
            elif x:
                find_label = r"{} \({}, .+\)".format(find_label, x)
            elif y:
                find_label = r"{} \(.+, {}\)".format(find_label, y)

            if re.search(find_label, label):
                selected_models[label] = models[label]

        return selected_models

    def _get_model_parameters(self, models=None, model_label=None, x=None, y=None):
        """
        Convert each parameter of model inside models into a coordinate that
        maps the model name and parameter name to a `astropy.units.Quantity`
        object.

        Parameters
        ----------
        models : dict
            A dictionary where the key is a model name and the value is an
            `astropy.modeling.CompoundModel` object.
        model_label : str
            Get model parameters for a particular model by inputting its label.
        x : int
            The x coordinate of the model spaxels that will be returned from
            get_models.
        y : int
            The y coordinate of the model spaxels that will be returned from
            get_models.

        Returns
        -------
        :dict: a dictionary of the form
            {model name: {parameter name: [[`astropy.units.Quantity`]]}}
            for 3d models or
            {model name: {parameter name: `astropy.units.Quantity`}} where the
            Quantity object represents the parameter value and unit of one of
            spaxel models or the 1d models, respectively.
        """
        if models and model_label:
            models = self.get_models(models=models, model_label=model_label, x=x, y=y)
        elif models is None and model_label:
            models = self.get_models(model_label=model_label, x=x, y=y)
        elif models is None:
            models = self.fitted_models

        data_shapes = {}
        for label in models:
            data_label = label.split(" (")[0]
            if data_label not in data_shapes:
                data_shapes[data_label] = self.app.data_collection[data_label].data.shape

        param_dict = {}
        parameters_cube = {}
        param_x_y = {}
        param_units = {}

        for label in models:
            # 3d models take the form of "Model (1,2)" so this if statement
            # looks for that style and separates out the pertinent information.
            if " (" in label:
                label_split = label.split(" (")
                model_name = label_split[0]
                x = int(label_split[1].split(", ")[0])
                y = int(label_split[1].split(", ")[1][:-1])

                # x and y values are added to this dict where they will be used
                # to convert the models of each spaxel into a single
                # coordinate in the parameters_cube dictionary.
                if model_name not in param_x_y:
                    param_x_y[model_name] = {'x': [], 'y': []}
                if x not in param_x_y[model_name]['x']:
                    param_x_y[model_name]['x'].append(x)
                if y not in param_x_y[model_name]['y']:
                    param_x_y[model_name]['y'].append(y)

            # 1d models will be handled by this else statement.
            else:
                model_name = label

            if model_name not in param_dict:
                param_dict[model_name] = list(models[label].param_names)

        # This adds another dictionary as the value of
        # parameters_cube[model_name] where the key is the parameter name
        # and the value is either a 2d array of zeros or a single value, depending
        # on whether the model in question is 3d or 1d, respectively.
        for model_name in param_dict:
            if model_name in param_x_y:
                parameters_cube[model_name] = {x: np.zeros(shape=data_shapes[model_name][:2])
                                               for x in param_dict[model_name]}
            else:
                parameters_cube[model_name] = {x: 0
                                               for x in param_dict[model_name]}

        # This loop handles the actual placement of param.values and
        # param.units into the parameter_cubes dictionary.
        for label in models:
            if " (" in label:
                label_split = label.split(" (")
                model_name = label_split[0]

                # If the get_models method is used to build a dictionary of
                # models and a value is set for the x or y parameters, that
                # will mean that only one x or y value is present in the
                # models.
                if len(param_x_y[model_name]['x']) == 1:
                    x = 0
                else:
                    x = int(label_split[1].split(", ")[0])

                if len(param_x_y[model_name]['y']) == 1:
                    y = 0
                else:
                    y = int(label_split[1].split(", ")[1][:-1])

                param_units[model_name] = {}

                for name in param_dict[model_name]:
                    param = getattr(models[label], name)
                    parameters_cube[model_name][name][x][y] = param.value
                    param_units[model_name][name] = param.unit
            else:
                model_name = label
                param_units[model_name] = {}

                # 1d models do not have anything set of param.unit, so the
                # return_units and input_units properties need to be used
                # instead, depending on the type of parameter `name` is.
                for name in param_dict[model_name]:
                    param = getattr(models[label], name)
                    parameters_cube[model_name][name] = param.value
                    param_units[model_name][name] = param.unit

        # Convert values of parameters_cube[key][param_name] into u.Quantity
        # objects that contain the appropriate unit set in
        # param_units[key][param_name]
        for key in parameters_cube:
            for param_name in parameters_cube[key]:
                parameters_cube[key][param_name] = u.Quantity(
                    parameters_cube[key][param_name],
                    param_units[key].get(param_name, None))

        return parameters_cube

    def get_model_parameters(self, model_label=None, x=None, y=None):
        """
        Convert each parameter of model inside models into a coordinate that
        maps the model name and parameter name to a `astropy.units.Quantity`
        object.

        Parameters
        ----------
        model_label : str
            Get model parameters for a particular model by inputting its label.
        x : int
            The x coordinate of the model spaxels that will be returned from
            get_models.
        y : int
            The y coordinate of the model spaxels that will be returned from
            get_models.

        Returns
        -------
        :dict: a dictionary of the form
            {model name: {parameter name: [[`astropy.units.Quantity`]]}}
            for 3d models or
            {model name: {parameter name: `astropy.units.Quantity`}} where the
            Quantity object represents the parameter value and unit of one of
            spaxel models or the 1d models, respectively.
        """
        return self._get_model_parameters(models=None, model_label=model_label,
                                          x=x, y=y)

    def vue_add_model(self, event):
        self.create_model_component()

    def vue_remove_model(self, event):
        self.remove_model_component(event)

    @observe('model_equation')
    def _check_model_equation_invalid(self, event=None):
        # Length is a dummy check to test the infrastructure
        if len(self.model_equation) == 0:
            self.model_equation_invalid_msg = 'model equation is required.'
            return
        if '' in self.equation_components:
            # includes an operator without a variable (ex: 'C+')
            self.model_equation_invalid_msg = 'incomplete equation.'
            return

        components_not_existing = [comp for comp in self.equation_components
                                   if comp not in self.model_components]
        if len(components_not_existing):
            if len(components_not_existing) == 1:
                msg = "is not an existing model component."
            else:
                msg = "are not existing model components."
            self.model_equation_invalid_msg = f'{", ".join(components_not_existing)} {msg}'
            return
        components_not_valid = [comp for comp in self.equation_components
                                if comp not in self.valid_model_components]
        if len(components_not_valid):
            if len(components_not_valid) == 1:
                msg = ("is currently disabled because it has"
                       " incompatible units with the current display units."
                       " Remove the component from the equation,"
                       " re-estimate its free parameters to use the new units"
                       " or revert the display units.")
            else:
                msg = ("are currently disabled because they have"
                       " incompatible units with the current display units."
                       " Remove the components from the equation,"
                       " re-estimate their free parameters to use the new units"
                       " or revert the display units.")
            self.model_equation_invalid_msg = f'{", ".join(components_not_valid)} {msg}'
            return
        self.model_equation_invalid_msg = ''

    @observe("dataset_selected", "dataset_items", "cube_fit")
    def _set_default_results_label(self, event={}):
        label_comps = []
        if hasattr(self, 'dataset') and (len(self.dataset.labels) > 1 or self.app.config == 'mosviz'):  # noqa
            label_comps += [self.dataset_selected]
        label_comps += ["model"]
        self.results_label_default = " ".join(label_comps)

    @observe("results_label")
    def _set_residuals_label_default(self, event={}):
        self.residuals_label_default = self.results_label+" residuals"

    @with_spinner()
    def calculate_fit(self, add_data=True):
        """
        Calculate the fit.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum/cube to the app as a data entry according to
            ``add_results``.

        Returns
        -------
        fitted model
        fitted spectrum/cube
        residuals (if ``residuals_calculate`` is set to ``True``)
        """

        if not self.spectral_subset_valid:
            valid, spec_range, subset_range = self._check_dataset_spectral_subset_valid(return_ranges=True)  # noqa
            raise ValueError(f"spectral subset '{self.spectral_subset.selected}' {subset_range} is outside data range of '{self.dataset.selected}' {spec_range}")  # noqa
        if len(self.model_equation_invalid_msg):
            raise ValueError(f"model equation is invalid: {self.model_equation_invalid_msg}")

        if self.cube_fit:
            ret = self._fit_model_to_cube(add_data=add_data)
        else:
            ret = self._fit_model_to_spectrum(add_data=add_data)

        if ret is None:  # pragma: no cover
            # something went wrong in the fitting call and (hopefully) already raised a warning,
            # but we don't have anything to add to the table
            return ret

        if self.cube_fit:
            # cube fits are currently unsupported in tables
            return ret

        row = {'model': self.results_label if add_data else '',
               'data_label': self.dataset_selected,
               'spectral_subset': self.spectral_subset_selected,
               'equation': self.equation.value}

        equation_components = self.equation_components
        for comp_ind, comp in enumerate(equation_components):
            for param_name, param_dict in self.get_model_component(comp).get('parameters', {}).items():  # noqa
                colprefix = f"{comp}:{param_name}_{comp_ind}"
                row[colprefix] = param_dict.get('value')
                row[f"{colprefix}:unit"] = param_dict.get('unit')
                row[f"{colprefix}:fixed"] = param_dict.get('fixed')
                row[f"{colprefix}:std"] = param_dict.get('std')

        self.table.add_item(row)

        return ret

    def vue_apply(self, event):
        self.calculate_fit()

    def _fit_model_to_spectrum(self, add_data):
        """
        Run fitting on the initialized models, fixing any parameters marked
        as such by the user, then update the displayed parameters with fit
        values
        """
        if self._warn_if_no_equation():
            return
        models_to_fit = self._reinitialize_with_fixed()

        spec = self.dataset.get_selected_spectrum(use_display_units=True)

        masked_spectrum = self._apply_subset_masks(spec,
                                                   self.spectral_subset)

        try:
            fitted_model, fitted_spectrum = fit_model_to_spectrum(
                masked_spectrum,
                models_to_fit,
                self.model_equation,
                run_fitter=True,
                window=None
            )
        except AttributeError:
            msg = SnackbarMessage("Unable to fit: model equation may be invalid",
                                  color="error", sender=self)
            self.hub.broadcast(msg)
            return

        selected_spec = self.dataset.selected_obj
        if '_pixel_scale_factor' in selected_spec.meta:
            fitted_spectrum.meta['_pixel_scale_factor'] = selected_spec.meta['_pixel_scale_factor']

        self._fitted_model = fitted_model
        self._fitted_spectrum = fitted_spectrum

        if add_data:
            self._fitted_models[self.results_label] = fitted_model
            self.add_results.add_results_from_plugin(fitted_spectrum)

            if self.residuals_calculate:
                # NOTE: this will NOT load into the viewer since we have already called
                # add_results_from_plugin above.
                self.add_results.add_results_from_plugin(masked_spectrum-fitted_spectrum,
                                                         label=self.residuals.value,
                                                         replace=False)

        self._set_default_results_label()

        # Update component model parameters with fitted values
        if isinstance(self._fitted_model, QuantityModel):
            self._update_parameters_from_QM()
        else:
            self._update_parameters_from_fit()

        # Also update the _initialized_models so we can use these values
        # as the starting point for cube fitting
        self._update_initialized_parameters()

        if self.residuals_calculate:
            return fitted_model, fitted_spectrum, masked_spectrum-fitted_spectrum
        return fitted_model, fitted_spectrum

    def _fit_model_to_cube(self, add_data):

        if self._warn_if_no_equation():
            return

        if self.dataset_selected in self.app.data_collection.labels:
            data = self.app.data_collection[self.dataset_selected]
        else:  # User selected some subset from spectrum viewer, just use original cube
            data = self.app.data_collection[0]

        # First, ensure that the selected data is cube-like. It is possible
        # that the user has selected a pre-existing 1d data object.
        if data.ndim != 3:
            snackbar_message = SnackbarMessage(
                f"Selected data {self.dataset_selected} is not cube-like",
                color='error',
                sender=self)
            self.hub.broadcast(snackbar_message)
            return

        # Get the primary data component
        if "_orig_spec" in data.meta:
            spec = data.meta["_orig_spec"]
        else:
            spec = data.get_object(cls=Spectrum1D, statistic=None)

        sb_unit = self.app._get_display_unit('sb')
        if spec.flux.unit != sb_unit:
            # ensure specutils has access to jdaviz custom unit equivalencies
            pixar_sr = spec.meta.get('_pixel_scale_factor', None)
            equivalencies = all_flux_unit_conversion_equivs(pixar_sr=pixar_sr,
                                                            cube_wave=spec.spectral_axis)

            pix2_in_flux = 'pix2' in spec.flux.unit.to_string()
            pix2_in_sb = 'pix2' in sb_unit
            # Handle various cases when PIX2 angle unit is present in the conversion
            if pix2_in_flux and pix2_in_sb:
                spec = spec.with_flux_unit(u.Unit(spec.flux.unit)*PIX2, equivalencies=equivalencies)  # noqa
                spec = spec.with_flux_unit(u.Unit(sb_unit)*PIX2, equivalencies=equivalencies)
            elif pix2_in_flux:
                spec = spec.with_flux_unit(u.Unit(spec.flux.unit) * PIX2, equivalencies=equivalencies)  # noqa
            elif pix2_in_sb:
                spec = spec.with_flux_unit(u.Unit(spec.flux.unit) / PIX2, equivalencies=equivalencies)  # noqa

            spec = spec.with_flux_unit(sb_unit, equivalencies=equivalencies)

        snackbar_message = SnackbarMessage(
            "Fitting model to cube...",
            loading=True, sender=self)
        self.hub.broadcast(snackbar_message)

        # Retrieve copy of the models with proper "fixed" dictionaries
        models_to_fit = self._reinitialize_with_fixed()

        # Apply masks from selected spectral subset
        spec = self._apply_subset_masks(spec, self.spectral_subset)
        # Also mask out NaNs for fitting. Simply adding filter_non_finite to the cube fit
        # didn't work out of the box, so doing this for now.
        if spec.mask is None:
            spec.mask = np.isnan(spec.flux)
        else:
            spec.mask = spec.mask | np.isnan(spec.flux)

        try:
            fitted_model, fitted_spectrum = fit_model_to_spectrum(
                spec,
                models_to_fit,
                self.model_equation,
                run_fitter=True,
                window=None
            )
        except ValueError:
            snackbar_message = SnackbarMessage(
                "Cube fitting failed",
                color='error', loading=False, sender=self)
            self.hub.broadcast(snackbar_message)
            raise

        # Save fitted 3D model in a way that the cubeviz
        # helper can access it.
        if add_data:
            for m in fitted_model:
                temp_label = "{} ({}, {})".format(self.results_label, m["x"], m["y"])
                self._fitted_models[temp_label] = m["model"]

        output_cube = Spectrum1D(flux=fitted_spectrum.flux, wcs=fitted_spectrum.wcs)

        selected_spec = self.dataset.selected_obj
        if '_pixel_scale_factor' in selected_spec.meta:
            output_cube.meta['_pixel_scale_factor'] = selected_spec.meta['_pixel_scale_factor']

        # Create new data entry for glue
        if add_data:
            self.add_results.add_results_from_plugin(output_cube)
            self._set_default_results_label()

        snackbar_message = SnackbarMessage(
            "Finished cube fitting",
            color='success', loading=False, sender=self)
        self.hub.broadcast(snackbar_message)

        return fitted_model, output_cube

    def _apply_subset_masks(self, spectrum, subset_component):
        """
        For a spectrum/spectral cube ``spectrum``, add a mask attribute
        if none exists. Mask excludes non-selected spectral subsets.
        """
        # only look for a mask if there is a selected subset:
        if subset_component.selected == subset_component.default_text:
            return spectrum

        spectrum = deepcopy(spectrum)
        subset_mask = subset_component.selected_subset_mask

        if spectrum.mask is not None:
            if subset_mask.ndim == 3:
                if spectrum.mask.ndim == 1:
                    # if subset mask is 3D and the `spectrum` mask is 1D, which
                    # happens when `spectrum` has been collapsed from 3D->1D,
                    # then also collapse the 3D mask in the spatial
                    # dimensions (0, 1) so that slices in the spectral axis that
                    # are masked in all pixels become masked in the spectral subset:
                    subset_mask = np.all(subset_mask, axis=(0, 1))
            spectrum.mask |= subset_mask
        else:
            if subset_mask.ndim < spectrum.flux.ndim:
                # correct the shape of spectral/spatial axes when they're different:
                subset_mask = np.broadcast_to(subset_mask, spectrum.flux.shape)

            elif (subset_mask.ndim == spectrum.flux.ndim and
                  subset_mask.shape != spectrum.flux.shape):
                # if the number of dimensions is correct but shape is
                # different, rearrange the arrays for specutils:
                subset_mask = np.swapaxes(subset_mask, 1, 0)

            spectrum.mask = subset_mask
        return spectrum



import os
import numpy as np
from glue.config import viewer_tool
from jdaviz.configs.default.plugins.tools import ProfileFromCube

__all__ = ['RampPerPixel']

ICON_DIR = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'data', 'icons')


@viewer_tool
class RampPerPixel(ProfileFromCube):

    # TODO: replace "pixelspectra" graphic with a "pixelramp" equivalent
    icon = os.path.join(ICON_DIR, 'pixelspectra.svg')
    tool_id = 'jdaviz:rampperpixel'
    action_text = 'See ramp at a single pixel'
    tool_tip = (
        'Click on the viewer and see the ramp profile '
        'at that pixel in the integration viewer'
    )

    def on_mouse_move(self, data):
        if data['event'] == 'mouseleave':
            self._mark.visible = False
            self._profile_viewer.reset_limits()
            return

        x = int(np.round(data['domain']['x']))
        y = int(np.round(data['domain']['y']))

        cube_cache = self.viewer.jdaviz_app._jdaviz_helper.cube_cache
        spectrum = cube_cache[list(cube_cache.keys())[0]].data

        if x >= spectrum.shape[0] or x < 0 or y >= spectrum.shape[1] or y < 0:
            self._mark.visible = False
        else:
            y_values = spectrum[x, y, :]
            if np.all(np.isnan(y_values)):
                self._mark.visible = False
                return
            self._mark.update_xy(np.arange(y_values.size), y_values)
            self._mark.visible = True
            self._profile_viewer.reset_limits()



from traitlets import Bool, List, Unicode, observe
import astropy.units as u

from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetSelectMixin,
                                        SpectralSubsetSelectMixin, with_spinner)
from jdaviz.core.user_api import PluginUserApi


__all__ = ['SonifyData']

try:
    import strauss  # noqa
    import sounddevice as sd
except ImportError:
    class Empty:
        pass
    sd = Empty()
    sd.default = Empty()
    sd.default.device = [-1, -1]
    _has_strauss = False
else:
    _has_strauss = True


@tray_registry('cubeviz-sonify-data', label="Sonify Data",
               viewer_requirements=['spectrum', 'image'])
class SonifyData(PluginTemplateMixin, DatasetSelectMixin, SpectralSubsetSelectMixin):
    """
    See the :ref:`Sonify Data Plugin Documentation <cubeviz-sonify-data>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    """
    template_file = __file__, "sonify_data.vue"

    # Removing UI option to vary these for now
    sample_rate = 44100  # IntHandleEmpty(44100).tag(sync=True)
    buffer_size = 2048  # IntHandleEmpty(2048).tag(sync=True)
    assidx = FloatHandleEmpty(2.5).tag(sync=True)
    ssvidx = FloatHandleEmpty(0.65).tag(sync=True)
    eln = Bool(True).tag(sync=True)
    audfrqmin = FloatHandleEmpty(50).tag(sync=True)
    audfrqmax = FloatHandleEmpty(1000).tag(sync=True)
    use_pccut = Bool(True).tag(sync=True)
    pccut = IntHandleEmpty(20).tag(sync=True)
    volume = IntHandleEmpty(100).tag(sync=True)
    stream_active = Bool(True).tag(sync=True)
    has_strauss = Bool(_has_strauss).tag(sync=True)

    # TODO: can we referesh the list, so sounddevices are up-to-date when dropdown clicked?
    sound_devices_items = List().tag(sync=True)
    sound_devices_selected = Unicode('').tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Sonify a data cube'
        self.docs_description = 'Sonify a data cube using the Strauss package.'
        if not self.has_strauss or sd.default.device[1] < 0:
            self.disabled_msg = ('To use Sonify Data, install strauss and restart Jdaviz. You '
                                 'can do this by running pip install strauss in the command'
                                 ' line and then launching Jdaviz. Currently, this plugin only'
                                 ' works on devices with valid sound output.')

        else:
            self.sound_device_indexes = None
            self.refresh_device_list()

        # TODO: Remove hardcoded range and flux viewer
        self.spec_viewer = self.app.get_viewer('spectrum-viewer')
        self.flux_viewer = self.app.get_viewer('flux-viewer')

    @property
    def user_api(self):
        expose = []
        return PluginUserApi(self, expose)

    @with_spinner()
    def vue_sonify_cube(self, *args):
        if self.disabled_msg:
            raise ValueError('Unable to sonify cube')

        # Get index of selected device
        selected_device_index = self.sound_device_indexes[self.sound_devices_selected]

        # Apply spectral subset bounds
        if self.spectral_subset_selected is not self.spectral_subset.default_text:
            display_unit = self.spec_viewer.state.x_display_unit
            min_wavelength = self.spectral_subset.selected_obj.lower.to_value(u.Unit(display_unit))
            max_wavelength = self.spectral_subset.selected_obj.upper.to_value(u.Unit(display_unit))
            self.flux_viewer.update_listener_wls((min_wavelength, max_wavelength), display_unit)

        # Ensure the current spectral region bounds are up-to-date at render time
        self.update_wavelength_range(None)
        # generate the sonified cube
        self.flux_viewer.get_sonified_cube(self.sample_rate, self.buffer_size,
                                           selected_device_index, self.assidx, self.ssvidx,
                                           self.pccut, self.audfrqmin,
                                           self.audfrqmax, self.eln, self.use_pccut)

        # Automatically select spectrum-at-spaxel tool
        spec_at_spaxel_tool = self.flux_viewer.toolbar.tools['jdaviz:spectrumperspaxel']
        self.flux_viewer.toolbar.active_tool = spec_at_spaxel_tool

    def vue_start_stop_stream(self, *args):
        self.stream_active = not self.stream_active
        self.flux_viewer.stream_active = not self.flux_viewer.stream_active

    @observe('spectral_subset_selected')
    def update_wavelength_range(self, event):
        if not hasattr(self, 'spec_viewer'):
            return
        display_unit = self.spec_viewer.state.x_display_unit
        # is this spectral selection or the entire spectrum?
        if hasattr(self.spectral_subset.selected_obj, "subregions"):
            wlranges = self.spectral_subset.selected_obj.subregions
        else:
            wlranges = None
        self.flux_viewer.update_listener_wls(wlranges, display_unit)

    @observe('volume')
    def update_volume_level(self, event):
        self.flux_viewer.update_volume_level(event['new'])

    @observe('sound_devices_selected')
    def update_sound_device(self, event):
        if event['new'] != event['old']:
            didx = dict(zip(*self.build_device_lists()))[event['new']]
            self.flux_viewer.update_sound_device(didx)

    def refresh_device_list(self):
        devices, indexes = self.build_device_lists()
        self.sound_device_indexes = dict(zip(devices, indexes))
        self.sound_devices_items = devices
        self.sound_devices_selected = dict(zip(indexes, devices))[sd.default.device[1]]

    def vue_refresh_device_list_in_dropdown(self, *args):
        self.refresh_device_list()

    def build_device_lists(self):
        # dedicated function to build the current *output*
        # device and index lists
        devices = []
        device_indexes = []
        for index, device in enumerate(sd.query_devices()):
            if device['max_output_channels'] > 0 and device['name'] not in devices:
                devices.append(device['name'])
                device_indexes.append(index)
        return devices, device_indexes



import os

import astropy.units as u
import pytest
from specutils import SpectralRegion

pytest.importorskip("strauss")
IN_GITHUB_ACTIONS = os.environ.get("CI", "false") == "true"


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test requires computer with audio output.")
def test_sonify_data(cubeviz_helper, spectrum1d_cube_larger):
    cubeviz_helper.load_data(spectrum1d_cube_larger, data_label="test")
    sonify_plg = cubeviz_helper.app.get_tray_item_from_name('cubeviz-sonify-data')
    assert sonify_plg.stream_active

    # Create sonified data cube
    sonify_plg.vue_sonify_cube()
    assert sonify_plg.flux_viewer.sonification_wl_bounds is None
    assert sonify_plg.flux_viewer.sonified_cube is not None

    # Test changing volume
    sonify_plg.volume = 90
    assert sonify_plg.flux_viewer.volume_level == 90

    # Test using spectral subset for setting sonification bounds
    spec_region = SpectralRegion(4.62360028e-07*u.m, 4.62920561e-07*u.m)
    subset_plugin = cubeviz_helper.plugins['Subset Tools']._obj
    subset_plugin.import_region(spec_region)
    sonify_plg.spectral_subset_selected = 'Subset 1'
    sonify_plg.vue_sonify_cube()
    assert sonify_plg.flux_viewer.sonification_wl_bounds == (4.62360028e-07, 4.62920561e-07)

    # Stop/start stream
    sonify_plg.vue_start_stop_stream()
    assert sonify_plg.flux_viewer.stream_active is False
    sonify_plg.vue_start_stop_stream()
    assert sonify_plg.flux_viewer.stream_active


@pytest.mark.skipif(not IN_GITHUB_ACTIONS, reason="Plugin disabled only in CI")
def test_sonify_data_disabled(cubeviz_helper, spectrum1d_cube_larger):
    cubeviz_helper.load_data(spectrum1d_cube_larger, data_label="test")
    sonify_plg = cubeviz_helper.app.get_tray_item_from_name('cubeviz-sonify-data')
    assert sonify_plg.disabled_msg
    with pytest.raises(ValueError, match='Unable to sonify cube'):
        sonify_plg.vue_sonify_cube()



"""
Tests the features of the Model Fitting Plugin (Selecting model parameters, adding models, etc.)
This does NOT test the actual fitting self (see test_fitting.py for that)
"""
import warnings
import pytest
from contextlib import nullcontext

import numpy as np
from numpy.testing import assert_allclose

from astropy.nddata import StdDevUncertainty
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.introspection import minversion
import astropy.units as u

from glue.core.roi import CircularROI
from specutils import Spectrum1D, SpectralRegion

from jdaviz.configs.default.plugins.model_fitting.initializers import MODELS

PYTEST_LT_8_0 = not minversion(pytest, "8.0.dev")


def test_default_model_labels(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']
    # By default, the spectral region should be the entire spectrum
    assert modelfit_plugin._obj.spectral_subset_selected == "Entire Spectrum"

    for model in MODELS:
        # Check that the auto label is set correctly (or at least the first character matches)
        # BlackBody and Polynomial labels behave differently, so check only the first character
        modelfit_plugin._obj.model_comp_selected = model
        assert modelfit_plugin._obj.comp_label[0] == model[0]

        # Test component label increments by default
        previous_label = modelfit_plugin._obj.comp_label
        modelfit_plugin._obj.vue_add_model({})
        assert modelfit_plugin._obj.comp_label == previous_label + "_1"

    assert len(modelfit_plugin._obj.component_models) == len(MODELS)

    # Test that default equation adds all components together
    assert (
        modelfit_plugin._obj.model_equation
        == "+".join(param["id"] for param in modelfit_plugin._obj.component_models)
    )


def test_custom_model_labels(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']

    for i, model in enumerate(MODELS):
        # Add one of each model with a unique name
        modelfit_plugin._obj.model_comp_selected = model
        modelfit_plugin._obj.comp_label = f"test_model_{i}"
        modelfit_plugin._obj.vue_add_model({})

    assert len(modelfit_plugin._obj.component_models) == len(MODELS)

    # Test that default equation adds all components together
    assert (
        modelfit_plugin._obj.model_equation
        == "+".join(param["id"] for param in modelfit_plugin._obj.component_models)
    )


def test_register_model_with_uncertainty_weighting(specviz_helper, spectrum1d):
    spectrum1d.uncertainty = StdDevUncertainty(spectrum1d.flux * 0.1)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']

    # Test registering a simple linear fit
    modelfit_plugin._obj.model_comp_selected = 'Linear1D'
    modelfit_plugin._obj.vue_add_model({})
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert len(specviz_helper.app.data_collection) == 2

    # Test fitting again overwrites original fit
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert len(specviz_helper.app.data_collection) == 2

    # Test custom model label
    test_label = 'my_Linear1D'
    modelfit_plugin._obj.results_label = test_label
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert test_label in specviz_helper.app.data_collection

    # Test that the parameter uncertainties were updated
    expected_uncertainties = {'slope': 0.0007063584243707317, 'intercept': 4.885421320000062}
    result_model = modelfit_plugin._obj.component_models[0]
    for param in result_model["parameters"]:
        assert np.allclose(param["std"], expected_uncertainties[param["name"]], rtol=0.01)


def test_register_model_uncertainty_is_none(specviz_helper, spectrum1d):
    # Set uncertainty to None
    spectrum1d.uncertainty = None
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']

    # Test registering a simple linear fit
    modelfit_plugin._obj.model_comp_selected = 'Linear1D'
    modelfit_plugin._obj.vue_add_model({})
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert len(specviz_helper.app.data_collection) == 2

    # Test fitting again overwrites original fit
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert len(specviz_helper.app.data_collection) == 2

    # Test custom model label
    test_label = 'my_Linear1D'
    modelfit_plugin._obj.results_label = test_label
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        modelfit_plugin.calculate_fit()
    assert test_label in specviz_helper.app.data_collection

    # Test that the parameter uncertainties were updated
    expected_uncertainties = {'slope': 0.00038, 'intercept': 2.67}
    result_model = modelfit_plugin._obj.component_models[0]
    for param in result_model["parameters"]:
        assert np.allclose(param["std"], expected_uncertainties[param["name"]], rtol=0.01)


def test_register_cube_model(cubeviz_helper, spectrum1d_cube):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        cubeviz_helper.load_data(spectrum1d_cube)
    modelfit_plugin = cubeviz_helper.plugins['Model Fitting']

    # Test custom model label
    modelfit_plugin.create_model_component('Linear1D', 'L')
    test_label = 'my_Linear1D'
    modelfit_plugin._obj.results_label = test_label
    # changing the lable should set auto to False, but the event may not have triggered yet
    modelfit_plugin._obj.results_label_auto = False
    modelfit_plugin.cube_fit = True
    modelfit_plugin.reestimate_model_parameters()
    assert modelfit_plugin._obj.results_label_default == 'model'
    assert modelfit_plugin._obj.results_label == test_label
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='.*Model is linear in parameters.*')
        modelfit_plugin.calculate_fit()
    assert test_label in cubeviz_helper.app.data_collection


def test_initialize_gaussian_with_cube(cubeviz_helper, spectrum1d_cube_larger):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        cubeviz_helper.load_data(spectrum1d_cube_larger)
    modelfit_plugin = cubeviz_helper.plugins['Model Fitting']

    modelfit_plugin.cube_fit = True
    # Check that we can initialize a Gaussian1D with cube fit toggled on
    modelfit_plugin.create_model_component('Gaussian1D', 'G')


def test_fit_cube_no_wcs(cubeviz_helper):
    # This is like when user does something to a cube outside of Jdaviz
    # and then loads it back into a new instance of Cubeviz for further analysis.
    sp = Spectrum1D(flux=np.ones((7, 8, 9)) * u.nJy)  # nx, ny, nz
    cubeviz_helper.load_data(sp, data_label="test_cube")
    mf = cubeviz_helper.plugins['Model Fitting']
    mf.create_model_component('Linear1D')
    mf.cube_fit = True
    # Need to manually reestimate the parameters to update the units
    mf.reestimate_model_parameters()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Model is linear in parameters.*")
        fitted_model, output_cube = mf.calculate_fit(add_data=True)
    assert len(fitted_model) == 56  # ny * nx
    # Make sure shapes are all self-consistent within Cubeviz instance.
    fitted_data = cubeviz_helper.app.data_collection["model"]
    assert fitted_data.shape == (7, 8, 9)  # nx, ny, nz
    assert fitted_data.shape == cubeviz_helper.app.data_collection[0].shape
    assert fitted_data.shape == output_cube.shape


def test_toggle_cube_fit_subset(cubeviz_helper):
    sp = Spectrum1D(flux=np.ones((7, 8, 9)) * u.nJy)  # ny, nx, nz
    cubeviz_helper.load_data(sp, data_label="test_cube")
    mf = cubeviz_helper.plugins['Model Fitting']

    unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    cubeviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(7.5 * unit,
                                                                        8 * unit))

    mf.spectral_subset = 'Subset 1'
    mf.cube_fit = True


def test_refit_plot_options(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']

    modelfit_plugin._obj.model_comp_selected = 'Const1D'
    modelfit_plugin._obj.comp_label = "C"
    modelfit_plugin._obj.vue_add_model({})

    with pytest.warns(AstropyUserWarning):
        modelfit_plugin.calculate_fit(add_data=True)

    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    atts = {"color": "red", "linewidth": 2, "alpha": 0.8}
    layer_state = [layer.state for layer in sv.layers if layer.layer.label == "model"][0]
    for att in atts:
        setattr(layer_state, att, atts[att])

    # Refit using the same name, which will replace the data by default.
    modelfit_plugin.create_model_component('Linear1D', 'L')

    with pytest.warns(AstropyUserWarning):
        modelfit_plugin.calculate_fit(add_data=True)

    layer_state = [layer.state for layer in sv.layers if layer.layer.label == "model"][0]

    for att in atts:
        assert atts[att] == getattr(layer_state, att)


def test_user_api(specviz_helper, spectrum1d):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        specviz_helper.load_data(spectrum1d)
    p = specviz_helper.plugins['Model Fitting']

    with pytest.raises(ValueError, match="blah is not a valid attribute and cannot be set"):
        p.blah = 5

    # even though the default label is set to C, adding Linear1D should default to its automatic
    # default label of 'L'
    assert p.model_component == 'Const1D'  # tests SelectPluginComponent's __eq__
    assert p.model_component_label.value == 'C'
    assert p.model_component_label == 'C'  # tests AutoTextField's __eq__
    p.create_model_component('Linear1D')
    assert p.model_components == ['L']

    with pytest.raises(ValueError, match='poly_order should only be passed if model_component is Polynomial1D'):  # noqa
        p.create_model_component('Linear1D', poly_order=2)

    with pytest.raises(ValueError, match="model component label 'L' already in use"):
        p.create_model_component('Linear1D', 'L')

    with pytest.raises(ValueError, match="model component with label 'dne' does not exist"):
        p.remove_model_component('dne')

    p.remove_model_component('L')
    assert len(p.model_components) == 0

    p.create_model_component('Polynomial1D', poly_order=2)
    assert p.model_components == ['P2']

    with pytest.raises(ValueError, match="'dne' is not a label of an existing model component"):
        p.get_model_component('dne')

    p.get_model_component('P2')
    p.set_model_component('P2', 'c0', value=0.2, fixed=True)
    assert p.get_model_component('P2', 'c0')['value'] == 0.2
    assert p.get_model_component('P2', 'c0')['fixed'] is True


def test_fit_gaussian_with_fixed_mean(specviz_helper, spectrum1d):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        specviz_helper.load_data(spectrum1d)
    modelfit_plugin = specviz_helper.plugins['Model Fitting']

    modelfit_plugin.create_model_component('Gaussian1D', 'G')
    params = modelfit_plugin._obj.component_models[0]['parameters']
    params[1]['fixed'] = True  # Fix mean

    old_amp = params[0]['value']
    old_mean = params[1]['value']
    old_std = params[2]['value']

    modelfit_plugin.residuals_calculate = True
    result, spectrum, resids = modelfit_plugin.calculate_fit()

    # Make sure mean is really fixed.
    assert_allclose(result.mean.value, old_mean)
    assert not np.allclose((result.amplitude.value, result.stddev.value), (old_amp, old_std))


def test_reestimate_parameters(specviz_helper, spectrum1d):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        specviz_helper.load_data(spectrum1d)
    mf = specviz_helper.plugins['Model Fitting']

    mf.create_model_component('Gaussian1D', 'G')
    mf.set_model_component('G', 'stddev', value=1, fixed=True)
    mc = mf.get_model_component('G')

    assert_allclose(mc['parameters']['mean']['value'], 7055.519926198364)
    assert mc['parameters']['stddev']['value'] == 1
    assert mc['parameters']['stddev']['fixed'] is True

    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(7500 * unit,
                                                                        8000 * unit))

    mf.spectral_subset = 'Subset 1'

    mf.reestimate_model_parameters()
    mc = mf.get_model_component('G')

    # mean should change, stddev should not
    assert_allclose(mc['parameters']['mean']['value'], 7780.8838070125375)
    assert mc['parameters']['stddev']['value'] == 1
    assert mc['parameters']['stddev']['fixed'] is True


def test_subset_masks(cubeviz_helper, spectrum1d_cube_larger):
    cubeviz_helper.load_data(spectrum1d_cube_larger)
    assert spectrum1d_cube_larger.mask is None

    # create a "Subset 1" entry in spatial dimension, selected "interactively"
    cubeviz_helper.plugins['Subset Tools'].import_region(CircularROI(0.5, 0.5, 1))

    # check that when no subset is selected, the spectral cube has no mask:
    p = cubeviz_helper.app.get_tray_item_from_name('g-model-fitting')

    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    # create a "Subset 2" entry in spectral dimension, selected "interactively"
    min_wavelength = 4625 * u.AA
    max_wavelength = spectrum1d_cube_larger.wavelength.max()

    # This toolbar selection allows the XRangeROI call below to create a
    # new subset, rather than replacing the previous subset:
    sv.toolbar_active_subset.selected = []

    # Now create the new spectral subset:
    cubeviz_helper.plugins['Subset Tools'].import_region(
        SpectralRegion(min_wavelength.to(u.m), max_wavelength.to(u.m)))
    assert "Subset 2" in p.spectral_subset.choices

    # Select the spectral subset
    p.spectral_subset_selected = "Subset 2"

    # Get the data object again (ensures mask == None)
    data = cubeviz_helper.app.data_collection[-1].get_object()
    subset = cubeviz_helper.app.data_collection[-1].get_subset_object(
        p.spectral_subset_selected, cls=Spectrum1D, statistic=None
    )
    masked_data = p._apply_subset_masks(data, p.spectral_subset)

    expected_spectral_mask = np.ones(data.flux.shape).astype(bool)
    expected_spectral_mask[3:] = False

    assert np.all(masked_data.mask == expected_spectral_mask)
    assert np.all(subset.mask == expected_spectral_mask)


def test_invalid_subset(specviz_helper, spectrum1d):
    # 6000-8000
    specviz_helper.load_data(spectrum1d, data_label="right_spectrum")

    # 5000-7000
    sp2 = Spectrum1D(spectral_axis=spectrum1d.spectral_axis - 1000*spectrum1d.spectral_axis.unit,
                     flux=spectrum1d.flux * 1.25)
    specviz_helper.load_data(sp2, data_label="left_spectrum")

    # apply subset that overlaps on left_spectrum, but not right_spectrum
    # NOTE: using a subset that overlaps the right_spectrum (reference) results in errors when
    # retrieving the subset (https://github.com/spacetelescope/jdaviz/issues/1868)
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(5000 * unit,
                                                                        6000 * unit))

    plugin = specviz_helper.plugins['Model Fitting']
    plugin.create_model_component('Linear1D')

    plugin.dataset = 'right_spectrum'
    assert plugin.dataset == 'right_spectrum'
    assert plugin.spectral_subset == 'Entire Spectrum'
    assert plugin._obj.spectral_subset_valid

    plugin.spectral_subset = 'Subset 1'
    assert not plugin._obj.spectral_subset_valid

    with pytest.raises(ValueError, match=r"spectral subset 'Subset 1' \(5000.0, 6000.0\) is outside data range of 'right_spectrum' \(6000.0, 8000.0\)"):  # noqa
        plugin.calculate_fit()

    plugin.dataset = 'left_spectrum'
    assert plugin._obj.spectral_subset_valid


def test_all_nan_uncert(specviz_helper):

    # test that if you have a fully finite data array, and a fully nan/inf
    # uncert array, that it is set to None and the fit proceeds (rather than
    # being filtered in the fitter, as would normally happen with nans)

    uncertainty = StdDevUncertainty([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan] * u.Jy)
    spec = Spectrum1D(flux=[1, 2, 3, 4, 5, 6]*u.Jy, uncertainty=uncertainty)
    specviz_helper.load_data(spec)

    plugin = specviz_helper.plugins['Model Fitting']
    plugin.create_model_component('Linear1D')

    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        plugin.calculate_fit()

    # check that slope and intercept are fit correctly
    plugin._obj.component_models[0]['parameters'][0]['value'] == 1.0
    plugin._obj.component_models[0]['parameters'][1]['value'] == 1.0

    # and that this value is correctly set to false, even though there IS
    # a mismatch, since its the entire array it will be reset
    assert plugin._obj.non_finite_uncertainty_mismatch is False


def test_all_nan_uncert_subset(specviz_helper):

    # test that nans in uncertainty array are filtered from fit (contrary to
    # what is tested in test_all_nan_uncert, when its not the entire array they
    # SHOULD be filtered even when corresponding data values are finite), and that
    # the `non_finite_uncertainty_mismatch` traitlet is True to trigger a warning
    # message

    uncertainty = StdDevUncertainty([1, 1, np.nan, np.nan, np.nan, np.nan] * u.Jy)
    spec = Spectrum1D(flux=[2, 4, 3, 4, 5, 6]*u.Jy, uncertainty=uncertainty)
    specviz_helper.load_data(spec)

    plugin = specviz_helper.plugins['Model Fitting']
    plugin.create_model_component('Linear1D')

    if PYTEST_LT_8_0:
        ctx = nullcontext()
    else:
        ctx = pytest.warns(
            AstropyUserWarning, match="Non-Finite input data has been removed by the fitter")

    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'), ctx:
        plugin.calculate_fit()

    # check that slope and intercept are fit correctly to just the first 2
    # data points
    plugin._obj.component_models[0]['parameters'][0]['value'] == 2.0
    plugin._obj.component_models[0]['parameters'][1]['value'] == 2.0

    # # and that this value is correctly set to false, even though there IS
    # # a mismatch, since its the entire array it will be reset
    assert plugin._obj.non_finite_uncertainty_mismatch is True



from functools import cached_property
from specutils import Spectrum1D, SpectrumList

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


__all__ = ['SpecutilsSpectrumParser', 'SpecutilsSpectrumListParser']


@loader_parser_registry('specutils.Spectrum')
class SpecutilsSpectrumParser(BaseParser):
    SpecutilsCls = Spectrum1D

    @property
    def is_valid(self):
        if self.app.config not in ('specviz', 'specviz2d'):
            # NOTE: temporary during deconfig process
            return False
        try:
            self.output
        except Exception:
            return False
        return True

    @cached_property
    def output(self):
        return self.SpecutilsCls.read(self.input)


@loader_parser_registry('specutils.SpectrumList')
class SpecutilsSpectrumListParser(SpecutilsSpectrumParser):
    SpecutilsCls = SpectrumList

    @property
    def is_valid(self):
        if self.app.config != 'specviz':
            # NOTE: temporary during deconfig process
            return False
        return super().is_valid and len(self.output) > 1



# Command-line interface for jdaviz

import inspect
import os
import pathlib

from jdaviz import __version__
from jdaviz.app import _verbosity_levels, ALL_JDAVIZ_CONFIGS
from jdaviz import configs

__all__ = ['main']

CONFIGS_DIR = str(pathlib.Path(inspect.getfile(configs)).parent)
JDAVIZ_DIR = pathlib.Path(__file__).parent.resolve()
DEFAULT_VERBOSITY = 'warning'
DEFAULT_HISTORY_VERBOSITY = 'info'


def main(filepaths=None, layout='default', instrument=None, browser='default',
         theme='auto', verbosity=DEFAULT_VERBOSITY, history_verbosity=DEFAULT_HISTORY_VERBOSITY,
         host='localhost', port=0, hotreload=False):
    """
    Start a Jdaviz application instance with data loaded from FILENAME.

    Parameters
    ----------
    filepaths : list of str, optional
        List of paths to the file(s) to be loaded into the Jdaviz application.
    layout : str, optional
        Optional specification for which configuration to use on startup.
    instrument : str, optional
        Specifies which instrument parser to use for Mosviz, if applicable.
    browser : str, optional
        Path to browser executable.
    theme : {'auto', 'light', 'dark'}
        Theme to use for application.
    verbosity : {'debug', 'info', 'warning', 'error'}
        Verbosity of the popup messages in the application.
    history_verbosity : {'debug', 'info', 'warning', 'error'}
        Verbosity of the history logger in the application.
    host : str, optional
        Host to bind the server to, default is 'localhost'.
    port : int, optional
        Port to bind the server to, default is 0 (finds an empty port).
    hotreload : bool
        Whether to enable hot-reloading of the UI (for development)
    """
    if filepaths:
        # Convert paths to posix string; windows paths are not JSON compliant
        file_list = [pathlib.Path(f).absolute().as_posix() for f in filepaths]
        if layout == "imviz":
            # Imviz multiloading should be done all at once for batch processing.
            # Imviz convention is a single string of consecutive, comma-separated file paths
            file_list = [','.join(file_list)]
    else:
        file_list = []

    if layout == '' and len(file_list) > 1:
        raise ValueError("'layout' argument is required when specifying multiple files")

    # Keep track of start directory in environment variable so that it can be
    # easily accessed e.g. in the file load dialog.
    os.environ['JDAVIZ_START_DIR'] = os.path.abspath('.')

    from jdaviz import solara
    solara.config = layout.capitalize()
    solara.data_list = file_list
    if layout == 'mosviz':
        solara.load_data_kwargs = {'instrument': instrument}
    solara.theme = theme
    solara.jdaviz_verbosity = verbosity
    solara.jdaviz_history_verbosity = history_verbosity
    run_solara(host=host, port=port, theme=theme, browser=browser, production=not hotreload)


def run_solara(host, port, theme, browser, production: bool = True):
    os.environ["SOLARA_APP"] = "jdaviz.solara"
    import solara.server.starlette
    import solara.server.settings
    solara.server.settings.theme.variant = theme
    solara.server.settings.theme.loader = "plain"
    solara.server.settings.main.mode = "production" if production else "development"

    server = solara.server.starlette.ServerStarlette(host="localhost", port=port)
    print(f"Starting server on {server.base_url}")
    server.serve_threaded()
    server.wait_until_serving()
    if browser == "qt":
        from . import qt
        qt.run_qt(server.base_url)
    else:
        import webbrowser
        controller = webbrowser.get(None if browser == 'default' else browser)
        controller.open(server.base_url)
        server.join()


def _main(config=None):
    import argparse
    import sys

    parser = argparse.ArgumentParser(description='Start a Jdaviz application instance with data '
                                     'loaded from FILENAME.')
    filepaths_nargs = '*'
    if config is None:
        parser.add_argument('--layout', default='', choices=ALL_JDAVIZ_CONFIGS,
                            help='Configuration to use.')
    if (config == "mosviz") or ("mosviz" in sys.argv):
        filepaths_nargs = 1
    parser.add_argument('filepaths', type=str, nargs=filepaths_nargs, default=None,
                        help='The paths to the files to be loaded into the Jdaviz application.')
    parser.add_argument('--instrument', type=str, default='nirspec',
                        help='Manually specifies which instrument parser to use, for Mosviz')
    parser.add_argument('--browser', type=str, default='default',
                        help='Browser to use for application (use qt for embedded Qt browser).')
    parser.add_argument('--theme', choices=['light', 'dark'], default='light',
                        help='Theme to use for application.')
    parser.add_argument('--verbosity', choices=_verbosity_levels, default='info',
                        help='Verbosity of the application for popup snackbars.')
    parser.add_argument('--history-verbosity', choices=_verbosity_levels, default='info',
                        help='Verbosity of the logger history.')
    parser.add_argument('--host', type=str, default='localhost',
                        help='Host to bind the server to, defaults to localhost.')
    parser.add_argument('--port', type=int, default=0,
                        help='Port to bind the server to, defaults to 0 (finds an empty port).')
    # Also enables --no-hotreload
    parser.add_argument('--hotreload', action=argparse.BooleanOptionalAction, default=False,
                        help='Whether to enable hot-reloading of the UI (for development).')
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    args = parser.parse_args()

    if config is None:
        layout = args.layout
    else:
        layout = config

    main(filepaths=args.filepaths, layout=layout, instrument=args.instrument, browser=args.browser,
         theme=args.theme, verbosity=args.verbosity, history_verbosity=args.history_verbosity,
         host=args.host, port=args.port, hotreload=args.hotreload)


def _specviz():
    _main(config='specviz')


def _specviz2d():
    _main(config='specviz2d')


def _imviz():
    _main(config='imviz')


def _cubeviz():
    _main(config='cubeviz')


def _mosviz():
    _main(config='mosviz')


if __name__ == "__main__":
    _main()



from glue.core import BaseData
from glue_jupyter.bqplot.image import BqplotImageView

from jdaviz.core.registries import viewer_registry
from jdaviz.configs.cubeviz.plugins.mixins import WithSliceIndicator, WithSliceSelection
from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin
from jdaviz.configs.specviz.plugins.viewers import SpecvizProfileView
from jdaviz.core.freezable_state import FreezableBqplotImageViewerState
from jdaviz.configs.cubeviz.plugins.cube_listener import CubeListenerData, MINVOL
import numpy as np
from astropy import units as u

try:
    import sounddevice as sd
except ImportError:
    pass

__all__ = ['CubevizImageView', 'CubevizProfileView']


@viewer_registry("cubeviz-image-viewer", label="Image 2D (Cubeviz)")
class CubevizImageView(JdavizViewerMixin, WithSliceSelection, BqplotImageView):
    # categories: zoom resets, (zoom, pan), subset, select tools, shortcuts
    # NOTE: zoom and pan are merged here for space consideration and to avoid
    # overflow to second row when opening the tray
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:pixelboxzoommatch', 'jdaviz:boxzoom',
                     'jdaviz:pixelpanzoommatch', 'jdaviz:panzoom'],
                    ['bqplot:truecircle', 'bqplot:rectangle', 'bqplot:ellipse',
                     'bqplot:circannulus'],
                    ['jdaviz:spectrumperspaxel'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = None
    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # provide reference from state back to viewer to use for zoom syncing
        self.state._set_viewer(self)

        self._subscribe_to_layers_update()
        self.state.add_callback('reference_data', self._initial_x_axis)

        # Hide axes by default
        self.state.show_axes = False

        self.sonified_cube = None
        self.stream = None

        self.sonification_wl_ranges = None
        self.sonification_wl_unit = None
        self.volume_level = None
        self.stream_active = True

        self.data_menu._obj.dataset.add_filter('is_cube_or_image')

    @property
    def _default_spectrum_viewer_reference_name(self):
        return self.jdaviz_helper._default_spectrum_viewer_reference_name

    @property
    def _default_flux_viewer_reference_name(self):
        return self.jdaviz_helper._default_flux_viewer_reference_name

    @property
    def _default_uncert_viewer_reference_name(self):
        return self.jdaviz_helper._default_uncert_viewer_reference_name

    def _initial_x_axis(self, *args):
        # Make sure that the x_att is correct on data load
        ref_data = self.state.reference_data
        if ref_data and ref_data.ndim == 3:
            for att_name in ["Right Ascension", "RA", "Galactic Longitude"]:
                if att_name in ref_data.component_ids():
                    x_att = att_name
                    self.state.x_att_world = ref_data.id[x_att]
                    break
            else:
                x_att = "Pixel Axis 0 [z]"
                self.state.x_att = ref_data.id[x_att]

    def set_plot_axes(self):
        self.figure.axes[1].tick_format = None
        self.figure.axes[0].tick_format = None

        self.figure.axes[1].label = "y: pixels"
        self.figure.axes[0].label = "x: pixels"

        # Make it so y axis label is not covering tick numbers.
        self.figure.axes[1].label_offset = "-50"

    def data(self, cls=None):
        return [layer_state.layer  # .get_object(cls=cls or self.default_class)
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                isinstance(layer_state.layer, BaseData)]

    def start_stream(self):
        if self.stream and not self.stream.closed and self.stream_active:
            self.stream.start()

    def stop_stream(self):
        if self.stream and not self.stream.closed and self.stream_active:
            self.stream.stop()

    def update_sonified_cube(self, x, y):
        if (not self.sonified_cube or not hasattr(self.sonified_cube, 'newsig') or
                not hasattr(self.sonified_cube, 'sigcube')):
            return
        self.sonified_cube.newsig = self.sonified_cube.sigcube[x, y, :]
        self.sonified_cube.cbuff = True

    def update_listener_wls(self, wranges, wunit):
        self.sonification_wl_ranges = wranges
        self.sonification_wl_unit = wunit

    def update_sound_device(self, device_index):
        if not self.sonified_cube:
            return

        self.stop_stream()
        self.stream = sd.OutputStream(samplerate=self.sample_rate, blocksize=self.buffer_size,
                                      device=device_index, channels=1, dtype='int16',
                                      latency='low', callback=self.sonified_cube.player_callback)

    def update_volume_level(self, level):
        if not self.sonified_cube:
            return
        self.volume_level = level
        self.sonified_cube.atten_level = int(1/np.clip((level/100.)**2, MINVOL, 1))

    def get_sonified_cube(self, sample_rate, buffer_size, device, assidx, ssvidx,
                          pccut, audfrqmin, audfrqmax, eln, use_pccut):
        spectrum = self.active_image_layer.layer.get_object(statistic=None)
        wlens = spectrum.wavelength.to('m').value
        flux = spectrum.flux.value
        self.sample_rate = sample_rate
        self.buffer_size = buffer_size

        if self.sonification_wl_ranges:
            wdx = np.zeros(wlens.size).astype(bool)
            for r in self.sonification_wl_ranges:
                # index just the spectral subregion
                wdx = np.logical_or(wdx,
                                    np.logical_and(wlens >= r[0].to_value(u.m),
                                                   wlens <= r[1].to_value(u.m)))
            wlens = wlens[wdx]
            flux = flux[:, :, wdx]

        pc_cube = np.percentile(np.nan_to_num(flux), np.clip(pccut, 0, 99), axis=-1)

        # clip zeros and remove NaNs
        clipped_arr = np.nan_to_num(np.clip(flux, 0, np.inf), copy=False)

        # make a rough white-light image from the clipped array
        whitelight = np.expand_dims(clipped_arr.sum(-1), axis=2)

        if use_pccut:
            # subtract any percentile cut
            clipped_arr -= np.expand_dims(pc_cube, axis=2)

            # and re-clip
            clipped_arr = np.clip(clipped_arr, 0, np.inf)

        self.sonified_cube = CubeListenerData(clipped_arr ** assidx, wlens, duration=0.8,
                                              samplerate=sample_rate, buffsize=buffer_size,
                                              wl_unit=self.sonification_wl_unit,
                                              audfrqmin=audfrqmin, audfrqmax=audfrqmax,
                                              eln=eln, vol=self.volume_level)
        self.sonified_cube.sonify_cube()
        self.sonified_cube.sigcube = (
                self.sonified_cube.sigcube * pow(whitelight / whitelight.max(),
                                                 ssvidx)).astype('int16')
        self.stream = sd.OutputStream(samplerate=sample_rate, blocksize=buffer_size, device=device,
                                      channels=1, dtype='int16', latency='low',
                                      callback=self.sonified_cube.player_callback)
        self.sonified_cube.cbuff = True


@viewer_registry("cubeviz-profile-viewer", label="Profile 1D (Cubeviz)")
class CubevizProfileView(SpecvizProfileView, WithSliceIndicator):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom', 'jdaviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['jdaviz:panzoom', 'jdaviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['bqplot:xrange'],
                    ['jdaviz:selectslice', 'jdaviz:selectline'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default_tool_priority', ['jdaviz:selectslice'])
        super().__init__(*args, **kwargs)

    @property
    def _default_flux_viewer_reference_name(self):
        return self.jdaviz_helper._default_flux_viewer_reference_name



import numpy as np
from astropy.nddata import NDDataArray
from glue.core import BaseData
from glue.core.subset import Subset
from glue_jupyter.bqplot.image import BqplotImageView

from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin, JdavizProfileView
from jdaviz.configs.cubeviz.plugins.mixins import WithSliceSelection, WithSliceIndicator
from jdaviz.core.registries import viewer_registry
from jdaviz.core.freezable_state import FreezableBqplotImageViewerState

__all__ = ['RampvizProfileView', 'RampvizImageView']


@viewer_registry("rampviz-profile-viewer", label="Profile 1D (Rampviz)")
class RampvizProfileView(JdavizProfileView, WithSliceIndicator):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom', 'jdaviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['jdaviz:panzoom', 'jdaviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['jdaviz:selectslice'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = NDDataArray
    _default_profile_subset_type = 'temporal'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default_tool_priority', ['jdaviz:selectslice'])
        super().__init__(*args, **kwargs)

    def _initialize_x_axis(self):
        if len(self.state.x_att_helper.choices):
            self.state.x_att = self.state.x_att_helper.choices[-1]
            self.set_plot_axes()
            self.reset_limits()

    def reset_limits(self):
        # override to reset to the global y limits including marks:
        global_y_min = np.inf
        global_y_max = -np.inf
        for mark in self.figure.marks:
            if len(mark.y) and mark.visible:
                global_y_min = min(global_y_min, np.nanmin(mark.y))
                global_y_max = max(global_y_max, np.nanmax(mark.y))

        for layer in self.state.layers:
            if not isinstance(layer.layer, Subset) and layer.visible:
                component = layer.layer.main_components[0]
                layer_y_min = layer.layer.get_component(component).data.min()
                layer_y_max = layer.layer.get_component(component).data.max()

                global_y_min = min(global_y_min, layer_y_min)
                global_y_max = max(global_y_max, layer_y_max)

        y_buffer = 0.1

        y_min = (1 - y_buffer) * global_y_min
        y_max = (1 + y_buffer) * global_y_max
        if y_min != self.state.y_min or y_max != self.state.y_max:
            self.set_limits(y_min=y_min, y_max=y_max)

    def set_plot_axes(self):

        with self.figure.hold_sync():
            self.figure.axes[0].label = "Group"
            self.figure.axes[1].label = self.state.y_display_unit

            # Make it so axis labels are not covering tick numbers.
            self.figure.fig_margin["left"] = 95
            self.figure.fig_margin["bottom"] = 60
            self.figure.send_state('fig_margin')  # Force update
            self.figure.axes[0].label_offset = "40"
            self.figure.axes[1].label_offset = "-70"
            # NOTE: with tick_style changed below, the default responsive ticks in bqplot result
            # in overlapping tick labels.  For now we'll hardcode at 8, but this could be removed
            # (default to None) if/when bqplot auto ticks react to styling options.
            self.figure.axes[1].num_ticks = 8

            # Set Y-axis to scientific notation
            self.figure.axes[1].tick_format = '0.1e'

            for i in (0, 1):
                self.figure.axes[i].tick_style = {'font-size': 15, 'font-weight': 600}


@viewer_registry("rampviz-image-viewer", label="Image 2D (Rampviz)")
class RampvizImageView(JdavizViewerMixin, WithSliceSelection, BqplotImageView):
    # categories: zoom resets, (zoom, pan), subset, select tools, shortcuts
    # NOTE: zoom and pan are merged here for space consideration and to avoid
    # overflow to second row when opening the tray
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:pixelboxzoommatch', 'jdaviz:boxzoom'],
                    ['jdaviz:pixelpanzoommatch', 'jdaviz:panzoom'],
                    ['bqplot:truecircle', 'bqplot:rectangle', 'bqplot:ellipse',
                     'bqplot:circannulus'],
                    ['jdaviz:rampperpixel'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = NDDataArray
    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # provide reference from state back to viewer to use for zoom syncing
        self.state._set_viewer(self)

        self._subscribe_to_layers_update()
        self.state.add_callback('reference_data', self._initial_x_axis)

        # Hide axes by default
        self.state.show_axes = False

    @property
    def _default_group_viewer_reference_name(self):
        return self.jdaviz_helper._default_group_viewer_reference_name

    @property
    def _default_diff_viewer_reference_name(self):
        return self.jdaviz_helper._default_diff_viewer_reference_name

    @property
    def _default_integration_viewer_reference_name(self):
        return self.jdaviz_helper._default_integration_viewer_reference_name

    def _initial_x_axis(self, *args):
        # Make sure that the x_att is correct on data load
        ref_data = self.state.reference_data

        if ref_data:
            if "Pixel Axis 0 [z]" in [comp.label for comp in ref_data.components]:
                self.state.x_att = ref_data.id["Pixel Axis 0 [z]"]
            else:
                self.state.x_att = ref_data.id["Pixel Axis 0 [y]"]

    def set_plot_axes(self):
        self.figure.axes[1].tick_format = None
        self.figure.axes[0].tick_format = None

        self.figure.axes[1].label = "y: pixels"
        self.figure.axes[0].label = "x: pixels"

        # Make it so y axis label is not covering tick numbers.
        self.figure.axes[1].label_offset = "-50"

    def data(self, cls=None):
        return [layer_state.layer  # .get_object(cls=cls or self.default_class)
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                isinstance(layer_state.layer, BaseData)]



import gwcs
import pytest
from astropy.modeling import models
from astropy.nddata import VarianceUncertainty
from astropy.tests.helper import assert_quantity_allclose
import astropy.units as u
from astropy.utils.data import download_file
import numpy as np
from packaging.version import Version
from specreduce import tracing, background, extract
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import SPEC_PHOTON_FLUX_DENSITY_UNITS

GWCS_LT_0_18_1 = Version(gwcs.__version__) < Version('0.18.1')


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore')
def test_plugin(specviz2d_helper):
    # TODO: Change back to smaller number (30?) when ITSD is convinced it is them and not us.
    #       Help desk ticket INC0183598, J. Quick.
    fn = download_file('https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits',
                       cache=True, timeout=100)

    specviz2d_helper.load_data(spectrum_2d=fn)

    pext = specviz2d_helper.app.get_tray_item_from_name('spectral-extraction')

    # test trace marks - won't be created until after opening the plugin
    sp2dv = specviz2d_helper.app.get_viewer('spectrum-2d-viewer')
    assert len(sp2dv.figure.marks) == 3

    pext.keep_active = True
    assert len(sp2dv.figure.marks) == 12
    assert pext.marks['trace'].visible is True
    assert len(pext.marks['trace'].x) > 0

    # create FlatTrace
    pext.trace_type_selected = 'Flat'
    pext.trace_pixel = 28
    trace = pext.export_trace(add_data=True)
    assert isinstance(trace, tracing.FlatTrace)
    assert trace.trace_pos == 28
    trace.trace_pos = 27
    pext.import_trace(trace)
    assert pext.trace_pixel == 27

    # offset existing trace
    pext.trace_trace_selected = 'trace'
    pext.trace_offset = 2
    trace = pext.export_trace(add_data=True)  # overwrite
    assert isinstance(trace, tracing.FlatTrace)

    # create FitTrace
    pext.trace_trace_selected = 'New Trace'
    pext.trace_type_selected = 'Polynomial'
    trace = pext.export_trace(add_data=True)
    assert isinstance(trace, tracing.FitTrace)
    assert trace.guess == 27
    trace = pext.export_trace(trace_pixel=26, add_data=False)
    assert trace.guess == 26
    trace.guess = 28
    pext.import_trace(trace)
    assert pext.trace_pixel == 28

    # offset existing trace
    pext.trace_trace_selected = 'trace'
    pext.trace_offset = 2
    trace = pext.export_trace(add_data=True)  # overwrite
    assert isinstance(trace, tracing.ArrayTrace)

    # TODO: Investigate extra hidden mark from glue-jupyter, see
    # https://github.com/spacetelescope/jdaviz/pull/2478#issuecomment-1731864411
    # 3 new trace objects should have been loaded and plotted in the spectrum-2d-viewer
    assert len(sp2dv.figure.marks) == 16

    # interact with background section, check marks
    pext.trace_trace_selected = 'New Trace'
    pext.trace_type_selected = 'Flat'
    pext.bg_separation = 5
    pext.bg_width = 3
    pext.bg_type_selected = 'TwoSided'
    for mark in ['bg1_center', 'bg2_center']:
        assert pext.marks[mark].visible is True
        assert len(pext.marks[mark].x) > 0
    bg = pext.export_bg()
    pext.import_bg(bg)
    assert pext.bg_type_selected == 'TwoSided'

    pext.bg_type_selected = 'Manual'
    assert len(pext.marks['bg1_center'].x) == 0
    assert len(pext.marks['bg2_center'].x) == 0
    assert len(pext.marks['bg1_lower'].x) > 0

    pext.bg_type_selected = 'OneSided'
    # only bg1 is populated for OneSided
    assert len(pext.marks['bg1_center'].x) > 0
    assert len(pext.marks['bg2_center'].x) == 0

    # create background image
    pext.bg_separation = 4
    bg = pext.export_bg()
    assert isinstance(bg, background.Background)
    assert len(bg.traces) == 1
    assert bg.traces[0].trace[0] == 28 + 4
    assert bg.width == 3
    bg = pext.export_bg(bg_width=3.3)
    assert bg.width == 3.3
    bg.width = 4
    pext.import_bg(bg)
    assert pext.bg_width == 4
    bg_img = pext.export_bg_img()
    assert isinstance(bg_img, Spectrum1D)
    bg_spec = pext.export_bg_spectrum()
    assert isinstance(bg_spec, Spectrum1D)
    bg_sub = pext.export_bg_sub()
    assert isinstance(bg_sub, Spectrum1D)

    # interact with extraction section, check marks
    pext.ext_width = 1
    for mark in ['bg1_center', 'bg2_center']:
        assert pext.marks[mark].visible is False
    for mark in ['ext_lower', 'ext_upper']:
        assert pext.marks[mark].visible is True
        assert len(pext.marks[mark].x) > 0

    # create subtracted spectrum
    ext = pext.export_extract(ext_width=3)
    assert isinstance(ext, extract.BoxcarExtract)
    assert ext.width == 3
    ext.width = 2
    pext.import_extract(ext)
    assert pext.ext_width == 2
    sp_ext = pext.export_extract_spectrum()
    assert isinstance(sp_ext, Spectrum1D)

    pext.ext_type_selected = 'Horne'
    sp_ext = pext.export_extract_spectrum()
    assert isinstance(sp_ext, Spectrum1D)

    # test API calls
    for step in ['trace', 'bg', 'ext']:
        pext.update_marks(step)

    # test exception handling
    pext.trace_type = 'Polynomial'
    pext.bg_type_selected = 'TwoSided'
    pext.bg_separation = 1
    pext.bg_width = 5
    assert len(pext.ext_specreduce_err) > 0
    pext.bg_results_label = 'should not be created'
    pext.vue_create_bg_img()
    assert 'should not be created' not in [d.label for d in specviz2d_helper.app.data_collection]

    with pytest.raises(ValueError):
        pext.export_extract(invalid_kwarg=5)


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore')
def test_user_api(specviz2d_helper):
    fn = download_file('https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits', cache=True)  # noqa

    specviz2d_helper.load_data(spectrum_2d=fn)

    pext = specviz2d_helper.plugins['Spectral Extraction']
    pext.keep_active = True

    # test that setting a string to an AddResults object redirects to the label
    pext.bg_sub_add_results = 'override label'
    assert pext.bg_sub_add_results.label == 'override label'
    pext.bg_sub_add_results.label = 'override label 2'
    assert "override label 2" in pext.bg_sub_add_results._obj.__repr__()
    assert "override label 2" in pext.bg_sub_add_results._obj.auto_label.__repr__()

    pext.export_bg_sub(add_data=True)
    assert 'override label 2' in pext.ext_dataset.choices

    # test setting auto label
    pext.bg_sub_add_results.auto = True


@pytest.mark.remote_data
@pytest.mark.skipif(GWCS_LT_0_18_1, reason='Needs GWCS 0.18.1 or later')
def test_spectrum_on_top(specviz2d_helper):
    fn = download_file('https://mast.stsci.edu/api/v0.1/Download/file/?uri=mast:jwst/product/jw01538-o161_s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits', cache=True)  # noqa

    specviz2d_helper.load_data(spectrum_2d=fn)

    pext = specviz2d_helper.app.get_tray_item_from_name('spectral-extraction')
    pext.trace_pixel = 14.2
    assert pext.bg_type_selected == 'OneSided'
    np.testing.assert_allclose(pext.bg_separation, 6)


@pytest.mark.filterwarnings('ignore')
def test_horne_extract_self_profile(specviz2d_helper):

    spec2d = np.zeros((40, 100))
    spec2dvar = np.ones((40, 100))

    for ii in range(spec2d.shape[1]):
        mgaus = models.Gaussian1D(amplitude=10,
                                  mean=(9.+(20/spec2d.shape[1])*ii),
                                  stddev=2)
        rg = np.arange(0, spec2d.shape[0], 1)
        gaus = mgaus(rg)
        spec2d[:, ii] = gaus

    wave = np.arange(0, spec2d.shape[1], 1)
    objectspec = Spectrum1D(spectral_axis=wave*u.m,
                            flux=spec2d*u.Jy,
                            uncertainty=VarianceUncertainty(spec2dvar*u.Jy*u.Jy))

    specviz2d_helper.load_data(objectspec)
    pext = specviz2d_helper.app.get_tray_item_from_name('spectral-extraction')

    trace_fit = tracing.FitTrace(objectspec,
                                 trace_model=models.Polynomial1D(degree=1),
                                 window=13, peak_method='gaussian', guess=20)
    pext.import_trace(trace_fit)

    pext.ext_type.selected = "Horne"
    pext.horne_ext_profile.selected = "Self (interpolated)"

    # check that correct defaults are set
    assert pext.self_prof_n_bins == 10
    assert pext.self_prof_interp_degree_x == 1
    assert pext.self_prof_interp_degree_y == 1

    sp_ext = pext.export_extract_spectrum()

    bg_sub = pext.export_bg_sub()

    extract_horne_interp = extract.HorneExtract(bg_sub, trace_fit,
                                                spatial_profile='interpolated_profile')

    assert_quantity_allclose(extract_horne_interp.spectrum.flux, sp_ext.flux)

    # now try changing from defaults
    pext.self_prof_n_bins = 5
    pext.self_prof_interp_degree_x = 2
    pext.self_prof_interp_degree_y = 2

    sp_ext = pext.export_extract_spectrum()
    bg_sub = pext.export_bg_sub()

    extract_horne_interp = extract.HorneExtract(bg_sub, trace_fit,
                                                spatial_profile={'name': 'interpolated_profile',
                                                                 'n_bins_interpolated_profile': 5,
                                                                 'interp_degree': (2, 2)})

    assert_quantity_allclose(extract_horne_interp.spectrum.flux, sp_ext.flux)

    # test that correct errors are raised
    pext.self_prof_n_bins = 0
    with pytest.raises(ValueError, match='must be greater than 0'):
        sp_ext = pext.export_extract_spectrum()

    pext.self_prof_n_bins = 1
    pext.self_prof_interp_degree_x = 0
    with pytest.raises(ValueError, match='must be greater than 0'):
        sp_ext = pext.export_extract_spectrum()

    pext.self_prof_interp_degree_x = 1
    pext.self_prof_interp_degree_y = 0
    with pytest.raises(ValueError, match='`self_prof_interp_degree_y` must be greater than 0.'):
        sp_ext = pext.export_extract_spectrum()


def test_spectral_extraction_flux_unit_conversions(specviz2d_helper, mos_spectrum2d):
    specviz2d_helper.load_data(mos_spectrum2d)

    uc = specviz2d_helper.plugins["Unit Conversion"]
    pext = specviz2d_helper.plugins['Spectral Extraction']

    for new_flux_unit in SPEC_PHOTON_FLUX_DENSITY_UNITS:
        # iterate through flux units verifying that selected object/spectrum is obtained using
        # display units
        uc.flux_unit.selected = new_flux_unit

        exported_bg = pext.export_bg()
        assert exported_bg.image._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_bg_img = pext.export_bg_img()
        assert exported_bg_img._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_bg_sub = pext.export_bg_sub()
        assert exported_bg_sub._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_extract_spectrum = pext.export_extract_spectrum()
        assert exported_extract_spectrum._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_extract = pext.export_extract()
        assert exported_extract.image._unit == specviz2d_helper.app._get_display_unit('flux')



from .importer import *  # noqa
from .spectrum1d import *  # noqa
from .spectrum2d import *  # noqa
from .spectrum2d_as_list import *  # noqa
from .subset import *  # noqa
from .trace import *  # noqa


import numpy as np
from astropy import units as u
from astropy.nddata import NDData
from numpy.testing import assert_allclose, assert_array_equal

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_NoWCS


class TestLineProfileXY(BaseImviz_WCS_NoWCS):
    def test_plugin_linked_by_pixel(self):
        """Go through plugin logic but does not check plot contents."""
        lp_plugin = self.imviz.plugins['Image Profiles (XY)']._obj
        lp_plugin.plugin_opened = True

        assert lp_plugin.viewer.labels == ['imviz-0']
        assert lp_plugin.viewer_selected == 'imviz-0'

        # Plot attempt with null X/Y should not crash but also no-op.
        assert 'line' not in lp_plugin.plot_across_x.layers
        lp_plugin.vue_draw_plot()
        assert not lp_plugin.plot_available

        # Mimic "l" key pressed.
        lp_plugin._on_viewer_key_event(self.viewer,
                                       {'event': 'keydown', 'key': 'l',
                                        'domain': {'x': 5.1, 'y': 5}})
        assert_allclose(lp_plugin.selected_x, 5.1)
        assert_allclose(lp_plugin.selected_y, 5)
        assert len(lp_plugin.plot_across_x.layers['line'].layer.data['x']) > 0
        assert len(lp_plugin.plot_across_y.layers['line'].layer.data['x']) > 0
        assert lp_plugin.plot_available

        # Add data with unit
        ndd = NDData(np.ones((10, 10)), unit=u.nJy)
        self.imviz.load_data(ndd, data_label='ndd', show_in_viewer=False)

        viewer_2 = self.imviz.create_image_viewer()
        self.imviz.app.add_data_to_viewer(viewer_2.reference_id, 'has_wcs[SCI,1]')
        self.imviz.app.add_data_to_viewer(viewer_2.reference_id, 'ndd[DATA]')

        # Blink also triggers viewer takeover and line profile redraw,
        # similar to the "l" key but without touching X and Y.
        viewer_2.blink_once()
        assert lp_plugin.viewer.labels == ['imviz-0', 'imviz-1']
        assert lp_plugin.viewer_selected == 'imviz-1'
        assert_allclose(lp_plugin.selected_x, 5.1)
        assert_allclose(lp_plugin.selected_y, 5)
        assert lp_plugin.plot_across_x.layers['line'].visible
        assert len(lp_plugin.plot_across_x.layers['line'].layer.data['x']) > 0
        assert len(lp_plugin.plot_across_y.layers['line'].layer.data['x']) > 0
        assert lp_plugin.plot_available

        # Wrong input resets plots without error.
        lp_plugin.selected_x = -1
        lp_plugin.vue_draw_plot()
        assert 'line' not in lp_plugin.plot_across_x.layers
        assert not lp_plugin.plot_available

        # Mimic manual GUI inputs.
        lp_plugin.selected_x = 1.1
        lp_plugin.selected_y = 9
        lp_plugin.viewer_selected = 'imviz-0'
        assert lp_plugin.plot_across_x.layers['line'].visible
        assert len(lp_plugin.plot_across_x.layers['line'].layer.data['x']) > 0
        assert len(lp_plugin.plot_across_y.layers['line'].layer.data['x']) > 0
        assert lp_plugin.plot_available

        # Nothing should update on "l" when plugin closed.
        lp_plugin.plugin_opened = False
        lp_plugin._on_viewer_key_event(self.viewer,
                                       {'event': 'keydown', 'key': 'l',
                                        'domain': {'x': 5.1, 'y': 5}})
        lp_plugin.selected_x = 1.1
        lp_plugin.selected_y = 9


def test_line_profile_with_nan(imviz_helper):
    arr = np.ones((10, 10))
    arr[5, 5] = np.nan
    imviz_helper.load_data(arr)

    lp_plugin = imviz_helper.plugins['Image Profiles (XY)']._obj
    lp_plugin.plugin_opened = True
    lp_plugin.selected_x = 5
    lp_plugin.selected_y = 5
    lp_plugin.vue_draw_plot()
    assert lp_plugin.plot_available

    # NaN still in data but rendered properly as gap.
    # Cannot check the gap stuff in CI but can make sure X-axis is populated properly etc.
    for lp_plot in (lp_plugin.plot_across_x, lp_plugin.plot_across_y):
        assert lp_plot.layers['line'].state.line_visible
        assert not np.all(np.isfinite(lp_plot.layers['line'].layer.data['y']))
        assert_array_equal(lp_plot.layers['line'].layer.data['x'], range(10))
        assert_allclose([lp_plot.layers['line'].state.viewer_state.x_min,
                         lp_plot.layers['line'].state.viewer_state.x_max,
                         lp_plot.layers['line'].state.viewer_state.y_min,
                         lp_plot.layers['line'].state.viewer_state.y_max],
                        [0, 9, 0.95, 1.05])



import numpy as np
import pytest
from astropy.io import fits
from astropy.nddata import NDData

from jdaviz.configs.default.plugins.metadata_viewer.metadata_viewer import MetadataViewer


def test_view_dict(imviz_helper):
    mv = MetadataViewer(app=imviz_helper.app)
    arr = np.zeros((2, 2), dtype=np.float32)
    ndd_1 = NDData(arr, meta={
        'EXTNAME': 'SCI', 'EXTVER': 1, 'BAR': 10.0, '_hidden': 'no show',
        'HISTORY': 'Hmm', '': 'Invalid', 'FOO': '', 'COMMENT': 'a test', 'BOZO': None})
    ndd_2 = NDData(arr, meta={
        'EXTNAME': 'ASDF', 'REF': {'bar': 10.0, 'foo': {'1': '', '2': [1, 2]}}})

    # MEF
    ndd_3 = fits.HDUList([fits.PrimaryHDU(), fits.ImageHDU(arr)])
    ndd_3[1].name = 'DATA'
    ndd_4 = fits.HDUList([fits.PrimaryHDU(), fits.ImageHDU(arr)])
    ndd_4[0].header['APERTURE'] = ('#TODO', 'Aperture')
    ndd_4[1].name = 'DATA'

    imviz_helper.load_data(ndd_1, data_label='has_simple_meta')
    imviz_helper.load_data(ndd_2, data_label='has_nested_meta')
    imviz_helper.load_data(ndd_3, data_label='has_primary')
    imviz_helper.load_data(ndd_4, data_label='has_primary_2')
    imviz_helper.load_data(arr, data_label='no_meta')
    assert mv.dataset.labels == ['has_simple_meta[DATA]', 'has_nested_meta[DATA]',
                                 'has_primary[DATA,1]', 'has_primary_2[DATA,1]', 'no_meta']

    mv.dataset_selected = 'has_simple_meta[DATA]'
    assert not mv.has_primary
    assert not mv.show_primary
    assert not mv.has_comments
    assert mv.has_metadata
    assert mv.metadata == [
        ('BAR', '10.0', ''), ('BOZO', 'None', ''), ('EXTNAME', 'SCI', ''),
        ('EXTVER', '1', ''), ('FOO', '', '')]

    mv.dataset_selected = 'has_nested_meta[DATA]'
    assert not mv.has_primary
    assert not mv.show_primary
    assert not mv.has_comments
    assert mv.has_metadata
    assert mv.metadata == [
        ('EXTNAME', 'ASDF', ''), ('REF.bar', '10.0', ''),
        ('REF.foo.1', '', ''), ('REF.foo.2.0', '1', ''), ('REF.foo.2.1', '2', '')]

    mv.dataset_selected = 'has_primary[DATA,1]'
    assert mv.has_primary
    assert not mv.show_primary
    assert mv.has_comments
    assert mv.has_metadata
    assert mv.metadata == [('BITPIX', '-32', 'array data type'),
                           ('EXTNAME', 'DATA', 'extension name'),
                           ('GCOUNT', '1', 'number of groups'),
                           ('NAXIS', '2', 'number of array dimensions'),
                           ('NAXIS1', '2', ''), ('NAXIS2', '2', ''),
                           ('PCOUNT', '0', 'number of parameters'),
                           ('XTENSION', 'IMAGE', 'Image extension')]
    mv.show_primary = True
    assert mv.metadata == [('BITPIX', '8', 'array data type'), ('EXTEND', 'True', ''),
                           ('NAXIS', '0', 'number of array dimensions'),
                           ('SIMPLE', 'True', 'conforms to FITS standard')]

    mv.dataset_selected = 'has_primary_2[DATA,1]'
    assert mv.show_primary  # Make sure it sticks if possible
    assert mv.has_comments
    assert mv.metadata == [('APERTURE', '#TODO', 'Aperture'),
                           ('BITPIX', '8', 'array data type'),
                           ('EXTEND', 'True', ''),
                           ('NAXIS', '0', 'number of array dimensions'),
                           ('SIMPLE', 'True', 'conforms to FITS standard')]
    assert mv.meta == {'APERTURE': {"description": 'Aperture', "value": '#TODO'},
                       'BITPIX': {"description": 'array data type', "value": '8'},
                       'EXTEND': {"description": '', "value": 'True'},
                       'NAXIS': {"description": 'number of array dimensions', "value": '0'},
                       'SIMPLE': {"description": 'conforms to FITS standard', "value": 'True'}}

    mv.dataset_selected = 'no_meta'
    assert not mv.has_primary
    assert not mv.show_primary
    assert not mv.has_comments
    assert not mv.has_metadata
    assert mv.metadata == []
    assert mv.meta == {}


def test_view_invalid(imviz_helper):
    mv = MetadataViewer(app=imviz_helper.app)
    assert mv.dataset.labels == []

    # Should not even happen but if it does, do not crash.
    with pytest.raises(ValueError):
        mv.dataset_selected = 'foo'
    assert mv.dataset_selected == ''
    assert not mv.has_primary
    assert not mv.show_primary
    assert not mv.has_comments
    assert not mv.has_metadata
    assert mv.metadata == []
    assert mv.meta == {}



import sys
# this avoids:
# ValueError: Key backend: 'module://matplotlib_inline.backend_inline' is not a valid value for backend; supported values are [...]
# Although not 100% why, it has two effects:
#  1. PyInstaller picks it up as a module to include
#  2. It registers the backend, maybe earlier than it would be otherwise
import matplotlib_inline
import matplotlib_inline.backend_inline

# We still see the above error on CI on jdaviz, and the PyInstaller
# output recommends the following:
import matplotlib
matplotlib.use("module://matplotlib_inline.backend_inline")
# since matplotlib 3.9 (see https://github.com/matplotlib/matplotlib/pull/27948),
# it seems that matplotlib_inline.backend_inline is an alias for inline
# so we make sure to communicate that to PyInstaller
matplotlib.use("inline")

import jdaviz.cli


if __name__ == "__main__":
    # should change this to _main, but now it doesn't need arguments
    args = sys.argv.copy()
    # change the browser to qt if not specified
    if "--browser" not in args:
        args.append("--browser")
        args.append("qt")
    sys.argv = args
    jdaviz.cli._main()



from specreduce.tracing import Trace


from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection


@loader_importer_registry('Trace')
class TraceImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    @property
    def is_valid(self):
        return (isinstance(self.input, Trace)
                and 'Spectral Extraction' in self.app._jdaviz_helper.plugins)

    @property
    def default_viewer(self):
        # returns the registry name of the default viewer
        # only used if `show_in_viewer=True` and no existing viewers can accept the data
        return 'mosviz-profile-2d-viewer'



import os
from pathlib import Path
import threading

import solara
import solara.lab
import ipygoldenlayout
import ipysplitpanes
import ipyvue

import jdaviz
from jdaviz.app import custom_components

config = None
data_list = []
load_data_kwargs = {}
jdaviz_verbosity = 'error'
jdaviz_history_verbosity = 'info'


@solara.lab.on_kernel_start
def on_kernel_start():
    print("Starting kernel", solara.get_kernel_id())
    # at import time, solara runs with a dummy kernel
    # we simply ignore that
    if "dummy" in solara.get_kernel_id():
        return

    def on_kernel_close():
        def exit_process():
            # sys.exit(0) does not work, it just throws an exception
            # this really makes the process exit
            os._exit(0)
        # give the kernel some time to close
        threading.Thread(target=exit_process).start()
    return on_kernel_close


@solara.component
def Page():
    solara.Title("Jdaviz")

    if config is None:
        solara.Text("No config defined")
        return

    ipysplitpanes.SplitPanes()
    ipygoldenlayout.GoldenLayout()
    for name, path in custom_components.items():
        ipyvue.register_component_from_file(None, name,
                                            os.path.join(os.path.dirname(jdaviz.__file__), path))

    ipyvue.register_component_from_file('g-viewer-tab', "container.vue", jdaviz.__file__)

    solara.Style(Path(__file__).parent / "solara.css")

    if config is None or not hasattr(jdaviz.configs, config):
        from jdaviz.core.launcher import Launcher
        launcher = Launcher(height='100vh', filepath=(data_list[0] if len(data_list) == 1 else ''))
        solara.display(launcher.main_with_launcher)
        return

    viz = getattr(jdaviz.configs, config)(verbosity=jdaviz_verbosity,
                                          history_verbosity=jdaviz_history_verbosity)
    for data in data_list:
        if config == 'mosviz':
            viz.load_data(directory=data, **load_data_kwargs)
        else:
            viz.load_data(data, **load_data_kwargs)

    solara.display(viz.app)



from ipyvuetify import VuetifyTemplate

__all__ = ['StyleWidget']


class StyleWidget(VuetifyTemplate):
    def __init__(self, template_file, *args, **kwargs):
        self.template_file = template_file
        super().__init__(*args, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# copied and modified from astropy: https://github.com/astropy/astropy/pull/12304
"""
Models that have physical origins.
"""
import warnings

import numpy as np

from astropy import constants as const
from astropy import units as u
from astropy.utils.exceptions import AstropyUserWarning
from astropy.modeling.core import Fittable1DModel
from astropy.modeling.parameters import Parameter

__all__ = ["BlackBody"]


class BlackBody(Fittable1DModel):
    """
    Blackbody model using the Planck function.

    Parameters
    ----------
    temperature : `~astropy.units.Quantity` ['temperature']
        Blackbody temperature.

    scale : float or `~astropy.units.Quantity` ['dimensionless']
        Scale factor

    output_units: `~astropy.units.CompositeUnit` or string
        Desired output units when evaluating.  If a unit object,
        must be equivalent to one of erg / (cm ** 2 * s * Hz * sr)
        or erg / (cm ** 2 * s * AA * sr) for surface brightness
        or erg / (cm ** 2 * s * Hz) or erg / (cm ** 2 * s * AA) for
        flux density (in which case the pi is included internally
        and should not be passed in `scale` - see notes below).
        If a string, can be one of 'SNU', 'SLAM', 'FNU', 'FLAM'
        which reference the units above, respectively.

    Notes
    -----

    Model formula:

        .. math:: B_{\\nu}(T) = A \\frac{2 h \\nu^{3} / c^{2}}{exp(h \\nu / k T) - 1}

    Deprecated support for non-dimensionless units in `scale`:

        If `scale` is passed with non-dimensionless units that are equivalent
        to one of the supported `output_units`, the unit is stripped and interpreted
        as `output_units` and the float value will account for the scale between
        the passed units and native `output_units`.  If `scale` includes flux units,
        the value will be divided by pi (since solid angle is included in the units),
        and re-added internally when returning results.  Note that this can result in
        ambiguous and unexpected results.  To avoid confusion, pass unitless
        `scale` and `output_units` separately.

    Examples
    --------
    >>> from astropy.modeling import models
    >>> from astropy import units as u
    >>> bb = models.BlackBody(temperature=5000*u.K)
    >>> bb(6000 * u.AA)  # doctest: +FLOAT_CMP
    <Quantity 1.53254685e-05 erg / (Hz s sr cm2)>

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt

        from astropy.modeling.models import BlackBody
        from astropy import units as u
        from astropy.visualization import quantity_support

        bb = BlackBody(temperature=5778*u.K)
        wav = np.arange(1000, 110000) * u.AA
        flux = bb(wav)

        with quantity_support():
            plt.figure()
            plt.semilogx(wav, flux)
            plt.axvline(bb.nu_max.to_value(u.AA, equivalencies=u.spectral()), ls='--')
            plt.show()
    """

    # We parametrize this model with a temperature and a scale.
    temperature = Parameter(default=5000.0, min=0, unit=u.K, description="Blackbody temperature")
    scale = Parameter(default=1.0, min=0, description="Scale factor")

    # We allow values without units to be passed when evaluating the model, and
    # in this case the input x values are assumed to be frequencies in Hz or wavelengths
    # in AA (depending on the choice of output_units).
    _input_units_allow_dimensionless = True

    # We enable the spectral equivalency by default for the spectral axis
    input_units_equivalencies = {'x': u.spectral()}

    # Store the native units returned by B_nu equation
    _native_units = u.erg / (u.cm ** 2 * u.s * u.Hz * u.sr)

    # Store the base native output units.  If scale is passed with units
    # equivalent to these, the dimensionless factor is converted to these
    # units.
    _native_output_units = {'FNU': u.erg / (u.cm ** 2 * u.s * u.Hz),
                            'FLAM': u.erg / (u.cm ** 2 * u.s * u.AA),
                            'SNU': u.erg / (u.cm ** 2 * u.s * u.Hz * u.sr),
                            'SLAM': u.erg / (u.cm ** 2 * u.s * u.AA * u.sr)}

    # Until unit-support on scale is removed, let's raise a warning for the ambiguous cases
    # when calling bolometric_flux
    _bolometric_flux_ambig_warn = False

    def __init__(self, *args, **kwargs):
        output_units = kwargs.pop('output_units', None)
        scale = kwargs.get('scale', None)

        # DEPRECATE: for now we'll continue to support scale with non-dimensionless units
        # by stripping the unit and applying that to output_units.
        if hasattr(scale, 'unit') and not scale.unit.is_equivalent(u.dimensionless_unscaled):
            if output_units is None:
                # NOTE: output_units will have to pass validation when assigned below
                # so we don't need to duplicate the checks here
                output_units = scale.unit

                # NOTE: if the scale units are equivalent (but not identical) to the
                # "native" output units, we want to convert the dimensionless scale
                # to account for the scaling between these units.  This is self-consistent
                # with the treatment of pi below for flux units, and addresses
                # https://github.com/astropy/astropy/issues/11547#issuecomment-823772098
                # but is a change in behavior from the previous treatment where the
                # float value was treated as the scale factor, ignoring units
                for native_output_unit in self._native_output_units.values():
                    if output_units.is_equivalent(native_output_unit):
                        scale = scale.to(native_output_unit)
                        break

                # If the scale had FNU or FLAM units, then the scale quantity INCLUDES
                # pi*u.sr.  We need to remove the pi when passing the dimensionless
                # scale as it will be re-added later in evaluate.
                if output_units.is_equivalent(self._native_units*u.sr, u.spectral_density(1*u.Hz)):
                    kwargs['scale'] = scale.value / np.pi
                else:
                    kwargs['scale'] = scale.value

                self._bolometric_flux_ambig_warn = True

            else:
                # Do not allow passing scale with unit AND output_units
                # (even though they may be the same)
                raise ValueError("cannot pass output_units and scale with units")

        self.output_units = output_units
        return super().__init__(*args, **kwargs)

    def evaluate(self, x, temperature, scale):
        """Evaluate the model.

        Parameters
        ----------
        x : float, `~numpy.ndarray`, or `~astropy.units.Quantity` ['frequency']
            Frequency at which to compute the blackbody. If no units are given,
            this defaults to Hz.

        temperature : float, `~numpy.ndarray`, or `~astropy.units.Quantity`
            Temperature of the blackbody. If no units are given, this defaults
            to Kelvin.

        scale : float, `~numpy.ndarray`, or `~astropy.units.Quantity` ['dimensionless']
            Desired scale for the blackbody.

        Returns
        -------
        y : number or ndarray
            Blackbody spectrum. The units are determined from the units of
            ``scale``.

        .. note::

            Use `numpy.errstate` to suppress Numpy warnings, if desired.

        .. warning::

            Output values might contain ``nan`` and ``inf``.

        Raises
        ------
        ValueError
            Invalid temperature.

        ZeroDivisionError
            Wavelength is zero (when converting to frequency).
        """
        if not isinstance(temperature, u.Quantity):
            in_temp = u.Quantity(temperature, u.K)
        else:
            in_temp = temperature

        if not isinstance(x, u.Quantity):
            # then we assume it has input_units which depends on the
            # requested output units (either Hz or AA)
            in_x = u.Quantity(x, self.input_units['x'])
        else:
            in_x = x

        # Convert to units for calculations, also force double precision
        with u.add_enabled_equivalencies(u.spectral() + u.temperature()):
            freq = u.Quantity(in_x, u.Hz, dtype=np.float64)
            temp = u.Quantity(in_temp, u.K)

        # Check if input values are physically possible
        if np.any(temp < 0):
            raise ValueError(f"Temperature should be positive: {temp}")
        if not np.all(np.isfinite(freq)) or np.any(freq <= 0):
            warnings.warn(
                "Input contains invalid wavelength/frequency value(s)",
                AstropyUserWarning,
            )

        log_boltz = const.h * freq / (const.k_B * temp)
        boltzm1 = np.expm1(log_boltz)

        # Calculate blackbody flux
        bb_nu = 2.0 * const.h * freq ** 3 / (const.c ** 2 * boltzm1) / u.sr

        if self.scale.unit is not None:
            # Will be dimensionless at this point, but may not be dimensionless_unscaled
            if not hasattr(scale, 'unit'):
                # during fitting, scale will be passed without units
                # but we still need to convert from the input dimensionless
                # to dimensionless unscaled
                scale = scale * self.scale.unit
            scale = scale.to(u.dimensionless_unscaled)

        if self.output_units.is_equivalent(self._native_units, u.spectral_density(freq)):
            # then requesting SNU or SLAM
            bb_unit = self.output_units
            scale = scale  # dimensionless
        else:
            # then requesting FNU or FLAM, which means the bb_unit
            # needs to be the requested output units without sr
            # and we'll then multiply by pi * u.sr separately
            bb_unit = self.output_units / u.sr
            scale = scale * np.pi * u.sr

        y = scale * bb_nu.to(bb_unit, u.spectral_density(freq))

        # If the temperature parameter has no unit, we should return a unitless
        # value. This occurs for instance during fitting, since we drop the
        # units temporarily.
        if hasattr(temperature, "unit"):
            return y
        return y.value

    @property
    def output_units(self):
        if self._output_units is None:
            # default to SNU if not passed (or set internally from units on scale)
            return u.erg / (u.cm ** 2 * u.s * u.Hz * u.sr)
        return self._output_units

    @output_units.setter
    def output_units(self, unit):
        """ Ensure `output_units` is valid."""
        if isinstance(unit, str):
            if unit in ['SNU', 'SLAM', 'FNU', 'FLAM']:
                # let's provide some convenience for passing these as strings
                unit = self._native_output_units.get(unit)
            else:
                try:
                    unit = u.Unit(unit)
                except ValueError:
                    # then the string wasn't a valid unit, we'll allow the error below to be raised
                    pass

        if unit is None:
            pass
        elif isinstance(unit, u.UnitBase):
            # support SNU, SLAM, FNU, FLAM output units (and equivalent)
            if not (unit.is_equivalent(self._native_units, u.spectral_density(1*u.AA)) or
                    unit.is_equivalent(self._native_units*u.sr, u.spectral_density(1*u.AA))):
                raise ValueError(f"output_units not in surface brightness or flux density: {unit}")
        else:
            raise ValueError("output_units must be of type Unit, None, "
                             "or one of 'SNU', 'SLAM', 'FNU', 'FLAM'")

        self._output_units = unit

    @property
    def input_units(self):
        # The input units are those of the 'x' value, which will depend on the
        # units compatible with the expected output units.
        output_units = self.output_units
        if (output_units.is_equivalent(self._native_output_units['SNU']) or
                output_units.is_equivalent(self._native_output_units['FNU'])):
            return {self.inputs[0]: u.Hz}
        else:
            return {self.inputs[0]: u.AA}

    def _parameter_units_for_data_units(self, inputs_unit, outputs_unit):
        return {"temperature": u.K}

    @property
    def bolometric_flux(self):
        """Bolometric flux."""
        if self.scale.unit is not None:
            # Will be dimensionless at this point, but may not be dimensionless_unscaled
            scale = self.scale.quantity.to(u.dimensionless_unscaled)
        else:
            scale = self.scale.value

        if self._bolometric_flux_ambig_warn:
            warnings.warn(
                f"scale was originally passed with units, "
                f"but being treated as unitless with value={scale}",
                AstropyUserWarning,
            )

        # bolometric flux in the native units of the planck function
        native_bolflux = (
            scale * const.sigma_sb * self.temperature ** 4 / np.pi
        )
        # return in more "astro" units
        return native_bolflux.to(u.erg / (u.cm ** 2 * u.s))

    @property
    def lambda_max(self):
        """Peak wavelength when the curve is expressed as power density."""
        return const.b_wien / self.temperature

    @property
    def nu_max(self):
        """Peak frequency when the curve is expressed as power density."""
        return 2.8214391 * const.k_B * self.temperature / const.h
